"""
execnet io initialization code

creates io instances used for gateway io
"""
import os
import sys

try:
    from execnet.gateway_base import Popen2IO, Message
except ImportError:
    from __main__ import Popen2IO, Message

class Popen2IOMaster(Popen2IO):
    def __init__(self, args, execmodel):
        self.popen = p = execmodel.PopenPiped(args)
        Popen2IO.__init__(self, p.stdin, p.stdout, execmodel=execmodel)

    def wait(self):
        try:
            return self.popen.wait()
        except OSError:
            pass  # subprocess probably dead already

    def kill(self):
        killpopen(self.popen)

def killpopen(popen):
    try:
        if hasattr(popen, 'kill'):
            popen.kill()
        else:
            killpid(popen.pid)
    except EnvironmentError:
        sys.stderr.write("ERROR killing: %s\n" %(sys.exc_info()[1]))
        sys.stderr.flush()

def killpid(pid):
    if hasattr(os, 'kill'):
        os.kill(pid, 15)
    elif sys.platform == "win32" or getattr(os, '_name', None) == 'nt':
        try:
            import ctypes
        except ImportError:
            import subprocess
            # T: treekill, F: Force
            cmd = ("taskkill /T /F /PID %d" %(pid)).split()
            ret = subprocess.call(cmd)
            if ret != 0:
                raise EnvironmentError("taskkill returned %r" %(ret,))
        else:
            PROCESS_TERMINATE = 1
            handle = ctypes.windll.kernel32.OpenProcess(
                        PROCESS_TERMINATE, False, pid)
            ctypes.windll.kernel32.TerminateProcess(handle, -1)
            ctypes.windll.kernel32.CloseHandle(handle)
    else:
        raise EnvironmentError("no method to kill %s" %(pid,))



popen_bootstrapline = "import sys;exec(eval(sys.stdin.readline()))"


def popen_args(spec):
    python = spec.python or sys.executable
    args = str(python).split(' ')
    args.append('-u')
    if spec is not None and spec.dont_write_bytecode:
        args.append("-B")
    # Slight gymnastics in ordering these arguments because CPython (as of
    # 2.7.1) ignores -B if you provide `python -c "something" -B`
    args.extend(['-c', popen_bootstrapline])
    return args


def ssh_args(spec):
    remotepython = spec.python or "python"
    args = ["ssh", "-C" ]
    if spec.ssh_config is not None:
        args.extend(['-F', str(spec.ssh_config)])

    args.extend(spec.ssh.split())
    remotecmd = '%s -c "%s"' % (remotepython, popen_bootstrapline)
    args.append(remotecmd)
    return args

def create_io(spec, execmodel):
    if spec.popen:
        args = popen_args(spec)
        return Popen2IOMaster(args, execmodel)
    if spec.ssh:
        args = ssh_args(spec)
        io = Popen2IOMaster(args, execmodel)
        io.remoteaddress = spec.ssh
        return io

#
# Proxy Gateway handling code
#
# master: proxy initiator
# forwarder: forwards between master and sub
# sub: sub process that is proxied to the initiator

RIO_KILL = 1
RIO_WAIT = 2
RIO_REMOTEADDRESS = 3
RIO_CLOSE_WRITE = 4

class ProxyIO(object):
    """ A Proxy IO object allows to instantiate a Gateway
    through another "via" gateway.  A master:ProxyIO object
    provides an IO object effectively connected to the sub
    via the forwarder.  To achieve this, master:ProxyIO interacts
    with forwarder:serve_proxy_io() which itself
    instantiates and interacts with the sub.
    """
    def __init__(self, proxy_channel, execmodel):
        # after exchanging the control channel we use proxy_channel
        # for messaging IO
        self.controlchan = proxy_channel.gateway.newchannel()
        proxy_channel.send(self.controlchan)
        self.iochan = proxy_channel
        self.iochan_file = self.iochan.makefile('r')
        self.execmodel = execmodel

    def read(self, nbytes):
        return self.iochan_file.read(nbytes)

    def write(self, data):
        return self.iochan.send(data)

    def _controll(self, event):
        self.controlchan.send(event)
        return self.controlchan.receive()

    def close_write(self):
        self._controll(RIO_CLOSE_WRITE)

    def kill(self):
        self._controll(RIO_KILL)

    def wait(self):
        return self._controll(RIO_WAIT)

    @property
    def remoteaddress(self):
        return self._controll(RIO_REMOTEADDRESS)

    def __repr__(self):
        return '<RemoteIO via %s>' % (self.iochan.gateway.id, )

class PseudoSpec:
    def __init__(self, vars):
        self.__dict__.update(vars)
    def __getattr__(self, name):
        return None

def serve_proxy_io(proxy_channelX):
    execmodel = proxy_channelX.gateway.execmodel
    _trace = proxy_channelX.gateway._trace
    tag = "serve_proxy_io:%s " % proxy_channelX.id
    def log(*msg):
        _trace(tag + msg[0], *msg[1:])
    spec = PseudoSpec(proxy_channelX.receive())
    # create sub IO object which we will proxy back to our proxy initiator
    sub_io = create_io(spec, execmodel)
    control_chan = proxy_channelX.receive()
    log("got control chan", control_chan)

    # read data from master, forward it to the sub
    # XXX writing might block, thus blocking the receiver thread
    def forward_to_sub(data):
        log("forward data to sub, size %s" % len(data))
        sub_io.write(data)
    proxy_channelX.setcallback(forward_to_sub)

    def controll(data):
        if data==RIO_WAIT:
            control_chan.send(sub_io.wait())
        elif data==RIO_KILL:
            control_chan.send(sub_io.kill())
        elif data==RIO_REMOTEADDRESS:
            control_chan.send(sub_io.remoteaddress)
        elif data==RIO_CLOSE_WRITE:
            control_chan.send(sub_io.close_write())
    control_chan.setcallback(controll)

    # write data to the master coming from the sub
    forward_to_master_file = proxy_channelX.makefile("w")

    # read bootstrap byte from sub, send it on to master
    log('reading bootstrap byte from sub', spec.id)
    initial = sub_io.read(1)
    assert initial == '1'.encode('ascii'), initial
    log('forwarding bootstrap byte from sub', spec.id)
    forward_to_master_file.write(initial)

    # enter message forwarding loop
    while True:
        try:
            message = Message.from_io(sub_io)
        except EOFError:
            log('EOF from sub, terminating proxying loop', spec.id)
            break
        message.to_io(forward_to_master_file)
    # proxy_channelX will be closed from remote_exec's finalization code

if __name__ == "__channelexec__":
    serve_proxy_io(channel) # noqa



import json
import logging
import re
import os
import time

from ceph_deploy import conf, exc, admin
from ceph_deploy.cliutil import priority
from ceph_deploy.util.help_formatters import ToggleRawTextHelpFormatter
from ceph_deploy.util import paths, net, files, packages, system
from ceph_deploy.lib import remoto
from ceph_deploy.new import new_mon_keyring
from ceph_deploy import hosts
from ceph_deploy.misc import mon_hosts
from ceph_deploy import gatherkeys


LOG = logging.getLogger(__name__)


def mon_status_check(conn, logger, hostname, args):
    """
    A direct check for JSON output on the monitor status.

    For newer versions of Ceph (dumpling and newer) a new mon_status command
    was added ( `ceph daemon mon mon_status` ) and should be revisited if the
    output changes as this check depends on that availability.

    """
    asok_path = paths.mon.asok(args.cluster, hostname)

    out, err, code = remoto.process.check(
        conn,
        [
            'ceph',
            '--cluster={cluster}'.format(cluster=args.cluster),
            '--admin-daemon',
            asok_path,
            'mon_status',
        ],
    )

    for line in err:
        logger.error(line)

    try:
        return json.loads(b''.join(out).decode('utf-8'))
    except ValueError:
        return {}


def catch_mon_errors(conn, logger, hostname, cfg, args):
    """
    Make sure we are able to catch up common mishaps with monitors
    and use that state of a monitor to determine what is missing
    and warn apropriately about it.
    """
    monmap = mon_status_check(conn, logger, hostname, args).get('monmap', {})
    mon_initial_members = get_mon_initial_members(args, _cfg=cfg)
    public_addr = cfg.safe_get('global', 'public_addr')
    public_network = cfg.safe_get('global', 'public_network')
    mon_in_monmap = [
        mon.get('name')
        for mon in monmap.get('mons', [{}])
        if mon.get('name') == hostname
    ]
    if mon_initial_members is None or not hostname in mon_initial_members:
            logger.warning('%s is not defined in `mon initial members`', hostname)
    if not mon_in_monmap:
        logger.warning('monitor %s does not exist in monmap', hostname)
        if not public_addr and not public_network:
            logger.warning('neither `public_addr` nor `public_network` keys are defined for monitors')
            logger.warning('monitors may not be able to form quorum')


def mon_status(conn, logger, hostname, args, silent=False):
    """
    run ``ceph daemon mon.`hostname` mon_status`` on the remote end and provide
    not only the output, but be able to return a boolean status of what is
    going on.
    ``False`` represents a monitor that is not doing OK even if it is up and
    running, while ``True`` would mean the monitor is up and running correctly.
    """
    mon = 'mon.%s' % hostname

    try:
        out = mon_status_check(conn, logger, hostname, args)
        if not out:
            logger.warning('monitor: %s, might not be running yet' % mon)
            return False

        if not silent:
            logger.debug('*'*80)
            logger.debug('status for monitor: %s' % mon)
            for line in json.dumps(out, indent=2, sort_keys=True).split('\n'):
                logger.debug(line)
            logger.debug('*'*80)
        if out['rank'] >= 0:
            logger.info('monitor: %s is running' % mon)
            return True
        if out['rank'] == -1 and out['state']:
            logger.info('monitor: %s is currently at the state of %s' % (mon, out['state']))
            return True
        logger.info('monitor: %s is not running' % mon)
        return False
    except RuntimeError:
        logger.info('monitor: %s is not running' % mon)
        return False


def keyring_parser(path):
    """
    This is a very, very, dumb parser that will look for `[entity]` sections
    and return a list of those sections. It is not possible to parse this with
    ConfigParser even though it is almost the same thing.

    Since this is only used to spit out warnings, it is OK to just be naive
    about the parsing.
    """
    sections = []
    with open(path) as keyring:
        lines = keyring.readlines()
        for line in lines:
            line = line.strip('\n')
            if line.startswith('[') and line.endswith(']'):
                sections.append(line.strip('[]'))
    return sections


def concatenate_keyrings(args):
    """
    A helper to collect all keyrings into a single blob that will be
    used to inject it to mons with ``--mkfs`` on remote nodes

    We require all keyring files to be concatenated to be in a directory
    to end with ``.keyring``.
    """
    keyring_path = os.path.abspath(args.keyrings)
    LOG.info('concatenating keyrings from %s' % keyring_path)
    LOG.info('to seed remote monitors')

    keyrings = [
        os.path.join(keyring_path, f) for f in os.listdir(keyring_path)
        if os.path.isfile(os.path.join(keyring_path, f)) and f.endswith('.keyring')
    ]

    contents = []
    seen_sections = {}

    if not keyrings:
        path_from_arg = os.path.abspath(args.keyrings)
        raise RuntimeError('could not find any keyrings in %s' % path_from_arg)

    for keyring in keyrings:
        path = os.path.abspath(keyring)

        for section in keyring_parser(path):
            if not seen_sections.get(section):
                seen_sections[section] = path
                LOG.info('adding entity "%s" from keyring %s' % (section, path))
                with open(path) as k:
                    contents.append(k.read())
            else:
                LOG.warning('will not add keyring: %s' % path)
                LOG.warning('entity "%s" from keyring %s is a duplicate' % (section, path))
                LOG.warning('already present in keyring: %s' % seen_sections[section])

    return ''.join(contents)


def mon_add(args):
    cfg = conf.ceph.load(args)

    # args.mon is a list with only one entry
    mon_host = args.mon[0]

    try:
        with open('{cluster}.mon.keyring'.format(cluster=args.cluster),
                  'rb') as f:
            monitor_keyring = f.read()
    except IOError:
        raise RuntimeError(
            'mon keyring not found; run \'new\' to create a new cluster'
        )

    LOG.info('ensuring configuration of new mon host: %s', mon_host)
    args.client = args.mon
    admin.admin(args)
    LOG.debug(
        'Adding mon to cluster %s, host %s',
        args.cluster,
        mon_host,
    )

    mon_section = 'mon.%s' % mon_host
    cfg_mon_addr = cfg.safe_get(mon_section, 'mon addr')

    if args.address:
        LOG.debug('using mon address via --address %s' % args.address)
        mon_ip = args.address
    elif cfg_mon_addr:
        LOG.debug('using mon address via configuration: %s' % cfg_mon_addr)
        mon_ip = cfg_mon_addr
    else:
        mon_ip = net.get_nonlocal_ip(mon_host)
        LOG.debug('using mon address by resolving host: %s' % mon_ip)

    try:
        LOG.debug('detecting platform for host %s ...', mon_host)
        distro = hosts.get(
            mon_host,
            username=args.username,
            callbacks=[packages.ceph_is_installed]
        )
        LOG.info('distro info: %s %s %s', distro.name, distro.release, distro.codename)
        rlogger = logging.getLogger(mon_host)

        # ensure remote hostname is good to go
        hostname_is_compatible(distro.conn, rlogger, mon_host)
        rlogger.debug('adding mon to %s', mon_host)
        args.address = mon_ip
        distro.mon.add(distro, args, monitor_keyring)

        # tell me the status of the deployed mon
        time.sleep(2)  # give some room to start
        catch_mon_errors(distro.conn, rlogger, mon_host, cfg, args)
        mon_status(distro.conn, rlogger, mon_host, args)
        distro.conn.exit()

    except RuntimeError as e:
        LOG.error(e)
        raise exc.GenericError('Failed to add monitor to host:  %s' % mon_host)


def mon_create(args):

    cfg = conf.ceph.load(args)
    if not args.mon:
        args.mon = get_mon_initial_members(args, error_on_empty=True, _cfg=cfg)

    if args.keyrings:
        monitor_keyring = concatenate_keyrings(args)
    else:
        keyring_path = '{cluster}.mon.keyring'.format(cluster=args.cluster)
        try:
            monitor_keyring = files.read_file(keyring_path)
        except IOError:
            LOG.warning('keyring (%s) not found, creating a new one' % keyring_path)
            new_mon_keyring(args)
            monitor_keyring = files.read_file(keyring_path)

    LOG.debug(
        'Deploying mon, cluster %s hosts %s',
        args.cluster,
        ' '.join(args.mon),
        )

    errors = 0
    for (name, host) in mon_hosts(args.mon):
        try:
            # TODO add_bootstrap_peer_hint
            LOG.debug('detecting platform for host %s ...', name)
            distro = hosts.get(
                host,
                username=args.username,
                callbacks=[packages.ceph_is_installed]
            )
            LOG.info('distro info: %s %s %s', distro.name, distro.release, distro.codename)
            rlogger = logging.getLogger(name)

            # ensure remote hostname is good to go
            hostname_is_compatible(distro.conn, rlogger, name)
            rlogger.debug('deploying mon to %s', name)
            distro.mon.create(distro, args, monitor_keyring)

            # tell me the status of the deployed mon
            time.sleep(2)  # give some room to start
            mon_status(distro.conn, rlogger, name, args)
            catch_mon_errors(distro.conn, rlogger, name, cfg, args)
            distro.conn.exit()

        except RuntimeError as e:
            LOG.error(e)
            errors += 1

    if errors:
        raise exc.GenericError('Failed to create %d monitors' % errors)


def hostname_is_compatible(conn, logger, provided_hostname):
    """
    Make sure that the host that we are connecting to has the same value as the
    `hostname` in the remote host, otherwise mons can fail not reaching quorum.
    """
    logger.debug('determining if provided host has same hostname in remote')
    remote_hostname = conn.remote_module.shortname()
    if remote_hostname == provided_hostname:
        return
    logger.warning('*'*80)
    logger.warning('provided hostname must match remote hostname')
    logger.warning('provided hostname: %s' % provided_hostname)
    logger.warning('remote hostname: %s' % remote_hostname)
    logger.warning('monitors may not reach quorum and create-keys will not complete')
    logger.warning('*'*80)


def destroy_mon(conn, cluster, hostname):
    import datetime
    import time
    retries = 5

    path = paths.mon.path(cluster, hostname)

    if conn.remote_module.path_exists(path):
        # remove from cluster
        remoto.process.run(
            conn,
            [
                'ceph',
                '--cluster={cluster}'.format(cluster=cluster),
                '-n', 'mon.',
                '-k', '{path}/keyring'.format(path=path),
                'mon',
                'remove',
                hostname,
            ],
            timeout=7,
        )

        # stop
        if conn.remote_module.path_exists(os.path.join(path, 'upstart')) or system.is_upstart(conn):
            status_args = [
                'initctl',
                'status',
                'ceph-mon',
                'cluster={cluster}'.format(cluster=cluster),
                'id={hostname}'.format(hostname=hostname),
            ]

        elif conn.remote_module.path_exists(os.path.join(path, 'sysvinit')):
            status_args = [
                'service',
                'ceph',
                'status',
                'mon.{hostname}'.format(hostname=hostname),
            ]
        elif system.is_systemd(conn):
            status_args = [
                'systemctl',
                'stop',
                'ceph-mon@{hostname}.service'.format(hostname=hostname),
            ]
        else:
            raise RuntimeError('could not detect a supported init system, cannot continue')

        while retries:
            conn.logger.info('polling the daemon to verify it stopped')
            if is_running(conn, status_args):
                time.sleep(5)
                retries -= 1
                if retries <= 0:
                    raise RuntimeError('ceph-mon deamon did not stop')
            else:
                break

        # archive old monitor directory
        fn = '{cluster}-{hostname}-{stamp}'.format(
            hostname=hostname,
            cluster=cluster,
            stamp=datetime.datetime.utcnow().strftime("%Y-%m-%dZ%H:%M:%S"),
            )

        remoto.process.run(
            conn,
            [
                'mkdir',
                '-p',
                '/var/lib/ceph/mon-removed',
            ],
        )

        conn.remote_module.make_mon_removed_dir(path, fn)


def mon_destroy(args):
    errors = 0
    for (name, host) in mon_hosts(args.mon):
        try:
            LOG.debug('Removing mon from %s', name)

            distro = hosts.get(
                host,
                username=args.username,
                callbacks=[packages.ceph_is_installed]
            )
            hostname = distro.conn.remote_module.shortname()

            destroy_mon(
                distro.conn,
                args.cluster,
                hostname,
            )
            distro.conn.exit()

        except RuntimeError as e:
            LOG.error(e)
            errors += 1

    if errors:
        raise exc.GenericError('Failed to destroy %d monitors' % errors)


def mon_create_initial(args):
    mon_initial_members = get_mon_initial_members(args, error_on_empty=True)

    # create them normally through mon_create
    args.mon = mon_initial_members
    mon_create(args)

    # make the sets to be able to compare late
    mon_in_quorum = set([])
    mon_members = set([host for host in mon_initial_members])

    for host in mon_initial_members:
        mon_name = 'mon.%s' % host
        LOG.info('processing monitor %s', mon_name)
        sleeps = [20, 20, 15, 10, 10, 5]
        tries = 5
        rlogger = logging.getLogger(host)
        distro = hosts.get(
            host,
            username=args.username,
            callbacks=[packages.ceph_is_installed]
        )

        while tries:
            status = mon_status_check(distro.conn, rlogger, host, args)
            has_reached_quorum = status.get('state', '') in ['peon', 'leader']
            if not has_reached_quorum:
                LOG.warning('%s monitor is not yet in quorum, tries left: %s' % (mon_name, tries))
                tries -= 1
                sleep_seconds = sleeps.pop()
                LOG.warning('waiting %s seconds before retrying', sleep_seconds)
                time.sleep(sleep_seconds)  # Magic number
            else:
                mon_in_quorum.add(host)
                LOG.info('%s monitor has reached quorum!', mon_name)
                break
        distro.conn.exit()

    if mon_in_quorum == mon_members:
        LOG.info('all initial monitors are running and have formed quorum')
        LOG.info('Running gatherkeys...')
        gatherkeys.gatherkeys(args)
    else:
        LOG.error('Some monitors have still not reached quorum:')
        for host in mon_members - mon_in_quorum:
            LOG.error('%s', host)
        raise SystemExit('cluster may not be in a healthy state')


def mon(args):
    if args.subcommand == 'create':
        mon_create(args)
    elif args.subcommand == 'add':
        mon_add(args)
    elif args.subcommand == 'destroy':
        mon_destroy(args)
    elif args.subcommand == 'create-initial':
        mon_create_initial(args)
    else:
        LOG.error('subcommand %s not implemented', args.subcommand)


@priority(30)
def make(parser):
    """
    Ceph MON Daemon management
    """
    parser.formatter_class = ToggleRawTextHelpFormatter

    mon_parser = parser.add_subparsers(dest='subcommand')
    mon_parser.required = True

    mon_add = mon_parser.add_parser(
        'add',
        help=('R|Add a monitor to an existing cluster:\n'
              '\tceph-deploy mon add node1\n'
              'Or:\n'
              '\tceph-deploy mon add --address 192.168.1.10 node1\n'
              'If the section for the monitor exists and defines a `mon addr` that\n'
              'will be used, otherwise it will fallback by resolving the hostname to an\n'
              'IP. If `--address` is used it will override all other options.')
    )
    mon_add.add_argument(
        '--address',
        nargs='?',
    )
    mon_add.add_argument(
        'mon',
        nargs=1,
    )

    mon_create = mon_parser.add_parser(
        'create',
        help=('R|Deploy monitors by specifying them like:\n'
              '\tceph-deploy mon create node1 node2 node3\n'
              'If no hosts are passed it will default to use the\n'
              '`mon initial members` defined in the configuration.')
    )
    mon_create.add_argument(
        '--keyrings',
        nargs='?',
        help='concatenate multiple keyrings to be seeded on new monitors',
    )
    mon_create.add_argument(
        'mon',
        nargs='*',
    )

    mon_create_initial = mon_parser.add_parser(
        'create-initial',
        help=('Will deploy for monitors defined in `mon initial members`, '
              'wait until they form quorum and then gatherkeys, reporting '
              'the monitor status along the process. If monitors don\'t form '
              'quorum the command will eventually time out.')
    )
    mon_create_initial.add_argument(
        '--keyrings',
        nargs='?',
        help='concatenate multiple keyrings to be seeded on new monitors',
    )

    mon_destroy = mon_parser.add_parser(
        'destroy',
        help='Completely remove Ceph MON from remote host(s)'
    )
    mon_destroy.add_argument(
        'mon',
        nargs='+',
    )

    parser.set_defaults(
        func=mon,
    )

#
# Helpers
#


def get_mon_initial_members(args, error_on_empty=False, _cfg=None):
    """
    Read the Ceph config file and return the value of mon_initial_members
    Optionally, a NeedHostError can be raised if the value is None.
    """
    if _cfg:
        cfg = _cfg
    else:
        cfg = conf.ceph.load(args)
    mon_initial_members = cfg.safe_get('global', 'mon_initial_members')
    if not mon_initial_members:
        if error_on_empty:
            raise exc.NeedHostError(
                'could not find `mon initial members` defined in ceph.conf'
            )
    else:
        mon_initial_members = re.split(r'[,\s]+', mon_initial_members)
    return mon_initial_members


def is_running(conn, args):
    """
    Run a command to check the status of a mon, return a boolean.

    We heavily depend on the format of the output, if that ever changes
    we need to modify this.
    Check daemon status for 3 times
    output of the status should be similar to::

        mon.mira094: running {"version":"0.61.5"}

    or when it fails::

        mon.mira094: dead {"version":"0.61.5"}
        mon.mira094: not running {"version":"0.61.5"}
    """
    stdout, stderr, _ = remoto.process.check(
        conn,
        args
    )
    result_string = b' '.join(stdout)
    for run_check in [b': running', b' start/running']:
        if run_check in result_string:
            return True
    return False



from ceph_deploy.util import constants

def url(key_type, protocol="https"):
    return "{protocol}://{url}{key_type}.asc".format(
        protocol=protocol,
        url=constants.gpg_key_base_url,
        key_type=key_type
    )



def uninstall(distro, purge=False):
    packages = [
        'ceph',
        'ceph-mds',
        'ceph-common',
        'ceph-fs-common',
        'radosgw',
        ]
    extra_remove_flags = []
    if purge:
        extra_remove_flags.append('--purge')
    distro.packager.remove(
        packages,
        extra_remove_flags=extra_remove_flags
    )



try:
    import configparser
except ImportError:
    import ConfigParser as configparser
import contextlib
import sys

from ceph_deploy import exc


class _TrimIndentFile(object):
    def __init__(self, fp):
        self.fp = fp

    def readline(self):
        line = self.fp.readline()
        return line.lstrip(' \t')

    def __iter__(self):
        return iter(self.readline, '')

class CephConf(configparser.RawConfigParser):
    def __init__(self, *args, **kwargs):
        if sys.version_info >= (3, 2):
            kwargs.setdefault('strict', False)
        # super() cannot be used with an old-style class
        configparser.RawConfigParser.__init__(self, *args, **kwargs)

    def optionxform(self, s):
        s = s.replace('_', ' ')
        s = '_'.join(s.split())
        return s

    def safe_get(self, section, key):
        """
        Attempt to get a configuration value from a certain section
        in a ``cfg`` object but returning None if not found. Avoids the need
        to be doing try/except {ConfigParser Exceptions} every time.
        """
        try:
            #Use full parent function so we can replace it in the class
            # if desired
            return configparser.RawConfigParser.get(self, section, key)
        except (configparser.NoSectionError,
                configparser.NoOptionError):
            return None


def parse(fp):
    cfg = CephConf()
    ifp = _TrimIndentFile(fp)
    cfg.readfp(ifp)
    return cfg


def load(args):
    """
    :param args: Will be used to infer the proper configuration name, or
    if args.ceph_conf is passed in, that will take precedence
    """
    path = args.ceph_conf or '{cluster}.conf'.format(cluster=args.cluster)

    try:
        f = open(path)
    except IOError as e:
        raise exc.ConfigError(
            "%s; has `ceph-deploy new` been run in this directory?" % e
        )
    else:
        with contextlib.closing(f):
            return parse(f)


def load_raw(args):
    """
    Read the actual file *as is* without parsing/modifiying it
    so that it can be written maintaining its same properties.

    :param args: Will be used to infer the proper configuration name
    :paran path: alternatively, use a path for any configuration file loading
    """
    path = args.ceph_conf or '{cluster}.conf'.format(cluster=args.cluster)
    try:
        with open(path) as ceph_conf:
            return ceph_conf.read()
    except (IOError, OSError) as e:
        raise exc.ConfigError(
            "%s; has `ceph-deploy new` been run in this directory?" % e
        )


def write_conf(cluster, conf, overwrite):
    """ write cluster configuration to /etc/ceph/{cluster}.conf """
    import os

    path = '/etc/ceph/{cluster}.conf'.format(cluster=cluster)
    tmp = '{path}.{pid}.tmp'.format(path=path, pid=os.getpid())

    if os.path.exists(path):
        with open(path) as f:
            old = f.read()
            if old != conf and not overwrite:
                raise RuntimeError('config file %s exists with different content; use --overwrite-conf to overwrite' % path)
    with open(tmp, 'w') as f:
        f.write(conf)
        f.flush()
        os.fsync(f)
    os.rename(tmp, path)



import os
try:
    from urllib.parse import urlparse
except ImportError:
    from urlparse import urlparse

from ceph_deploy.lib import remoto
from ceph_deploy.util import templates


class PackageManager(object):
    """
    Base class for all Package Managers
    """

    def __init__(self, remote_conn):
        self.remote_info = remote_conn
        self.remote_conn = remote_conn.conn

    def _run(self, cmd, **kw):
        return remoto.process.run(
            self.remote_conn,
            cmd,
            **kw
        )

    def _check(self, cmd, **kw):
        return remoto.process.check(
            self.remote_conn,
            cmd,
            **kw
        )

    def install(self, packages, **kw):
        """Install packages on remote node"""
        raise NotImplementedError()

    def remove(self, packages, **kw):
        """Uninstall packages on remote node"""
        raise NotImplementedError()

    def clean(self):
        """Clean metadata/cache"""
        raise NotImplementedError()

    def add_repo_gpg_key(self, url):
        """Add given GPG key for repo verification"""
        raise NotImplementedError()

    def add_repo(self, name, url, **kw):
        """Add/rewrite a repo file"""
        raise NotImplementedError()

    def remove_repo(self, name):
        """Remove a repo definition"""
        raise NotImplementedError()


class RPMManagerBase(PackageManager):
    """
    Base class to hold common pieces of Yum and DNF
    """

    executable = None
    name = None

    def install(self, packages, **kw):
        if isinstance(packages, str):
            packages = [packages]

        extra_flags = kw.pop('extra_install_flags', None)
        cmd = [
            self.executable,
            '-y',
            'install',
        ]
        if extra_flags:
            if isinstance(extra_flags, str):
                extra_flags = [extra_flags]
            cmd.extend(extra_flags)

        cmd.extend(packages)
        return self._run(cmd)

    def remove(self, packages, **kw):
        if isinstance(packages, str):
            packages = [packages]

        extra_flags = kw.pop('extra_remove_flags', None)
        cmd = [
            self.executable,
            '-y',
            '-q',
            'remove',
        ]
        if extra_flags:
            if isinstance(extra_flags, str):
                extra_flags = [extra_flags]
            cmd.extend(extra_flags)
        cmd.extend(packages)
        return self._run(cmd)

    def clean(self, item=None):
        item = item or 'all'
        cmd = [
            self.executable,
            'clean',
            item,
        ]

        return self._run(cmd)

    def add_repo_gpg_key(self, url):
        cmd = ['rpm', '--import', url]
        self._run(cmd)

    def add_repo(self, name, url, **kw):
        gpg_url = kw.pop('gpg_url', None)
        if gpg_url:
            self.add_repo_gpg_key(gpg_url)
            gpgcheck=1
        else:
            gpgcheck=0

        # RPM repo defaults
        description = kw.pop('description', '%s repo' % name)
        enabled = kw.pop('enabled', 1)
        proxy = kw.pop('proxy', '') # will get ignored if empty
        _type = 'repo-md'
        baseurl = url.strip('/')  # Remove trailing slashes

        ceph_repo_content = templates.custom_repo(
            reponame=name,
            name=description,
            baseurl=baseurl,
            enabled=enabled,
            gpgcheck=gpgcheck,
            _type=_type,
            gpgkey=gpg_url,
            proxy=proxy,
            **kw
        )

        self.remote_conn.remote_module.write_yum_repo(
            ceph_repo_content,
            '%s.repo' % name
        )

    def remove_repo(self, name):
        filename = os.path.join(
            '/etc/yum.repos.d',
            '%s.repo' % name
        )
        self.remote_conn.remote_module.unlink(filename)


class DNF(RPMManagerBase):
    """
    The DNF Package manager
    """

    executable = 'dnf'
    name = 'dnf'

    def install(self, packages, **kw):
        extra_install_flags = kw.pop('extra_install_flags', [])
        if '--best' not in extra_install_flags:
            extra_install_flags.append('--best')
        super(DNF, self).install(
            packages,
            extra_install_flags=extra_install_flags,
            **kw
        )


class Yum(RPMManagerBase):
    """
    The Yum Package manager
    """

    executable = 'yum'
    name = 'yum'


class Apt(PackageManager):
    """
    Apt package management
    """

    executable = [
        'env',
        'DEBIAN_FRONTEND=noninteractive',
        'DEBIAN_PRIORITY=critical',
        'apt-get',
        '--assume-yes',
        '-q',
    ]
    name = 'apt'

    def install(self, packages, **kw):
        if isinstance(packages, str):
            packages = [packages]

        extra_flags = kw.pop('extra_install_flags', None)
        cmd = self.executable + [
            '--no-install-recommends',
            'install'
        ]

        if extra_flags:
            if isinstance(extra_flags, str):
                extra_flags = [extra_flags]
            cmd.extend(extra_flags)
        cmd.extend(packages)
        return self._run(cmd)

    def remove(self, packages, **kw):
        if isinstance(packages, str):
            packages = [packages]

        extra_flags = kw.pop('extra_remove_flags', None)
        cmd = self.executable + [
            '-f',
            '--force-yes',
            'remove'
        ]
        if extra_flags:
            if isinstance(extra_flags, str):
                extra_flags = [extra_flags]
            cmd.extend(extra_flags)

        cmd.extend(packages)
        return self._run(cmd)

    def clean(self):
        cmd = self.executable + ['update']
        return self._run(cmd)

    def add_repo_gpg_key(self, url):
        gpg_path = url.split('file://')[-1]
        if not url.startswith('file://'):
            cmd = ['wget', '-O', 'release.asc', url ]
            self._run(cmd, stop_on_nonzero=False)
        gpg_file = 'release.asc' if not url.startswith('file://') else gpg_path
        cmd = ['apt-key', 'add', gpg_file]
        self._run(cmd)

    def add_repo(self, name, url, **kw):
        gpg_url = kw.pop('gpg_url', None)
        if gpg_url:
            self.add_repo_gpg_key(gpg_url)

        safe_filename = '%s.list' % name.replace(' ', '-')
        mode = 0o644
        if urlparse(url).password:
            mode = 0o600
            self.remote_conn.logger.info(
                "Creating repo file with mode 0600 due to presence of password"
            )
        self.remote_conn.remote_module.write_sources_list(
            url,
            self.remote_info.codename,
            safe_filename,
            mode
        )

        # Add package pinning for this repo
        fqdn = urlparse(url).hostname
        self.remote_conn.remote_module.set_apt_priority(fqdn)

    def remove_repo(self, name):
        safe_filename = '%s.list' % name.replace(' ', '-')
        filename = os.path.join(
            '/etc/apt/sources.list.d',
            safe_filename
        )
        self.remote_conn.remote_module.unlink(filename)


class Zypper(PackageManager):
    """
    Zypper package management
    """

    executable = [
        'zypper',
        '--non-interactive',
        '--quiet'
    ]
    name = 'zypper'

    def install(self, packages, **kw):
        if isinstance(packages, str):
            packages = [packages]

        extra_flags = kw.pop('extra_install_flags', None)
        cmd = self.executable + ['install']
        if extra_flags:
            if isinstance(extra_flags, str):
                extra_flags = [extra_flags]
            cmd.extend(extra_flags)
        cmd.extend(packages)
        return self._run(cmd)

    def remove(self, packages, **kw):
        if isinstance(packages, str):
            packages = [packages]

        extra_flags = kw.pop('extra_remove_flags', None)
        cmd = self.executable + ['--ignore-unknown', 'remove']
        if extra_flags:
            if isinstance(extra_flags, str):
                extra_flags = [extra_flags]
            cmd.extend(extra_flags)
        cmd.extend(packages)
        stdout, stderr, exitrc = self._check(
            cmd,
            **kw
        )
        # exitrc is 104 when package(s) not installed.
        if not exitrc in [0, 104]:
            raise RuntimeError("Failed to execute command: %s" % " ".join(cmd))
        return

    def clean(self):
        cmd = self.executable + ['refresh']
        return self._run(cmd)



import argparse
import json
import logging
import os
import re
import sys
import time
from textwrap import dedent

from ceph_deploy import conf, exc, hosts, mon
from ceph_deploy.util import constants, system, packages
from ceph_deploy.cliutil import priority
from ceph_deploy.lib import remoto


LOG = logging.getLogger(__name__)


def get_bootstrap_osd_key(cluster):
    """
    Read the bootstrap-osd key for `cluster`.
    """
    path = '{cluster}.bootstrap-osd.keyring'.format(cluster=cluster)
    try:
        with open(path, 'rb') as f:
            return f.read()
    except IOError:
        raise RuntimeError('bootstrap-osd keyring not found; run \'gatherkeys\'')


def create_osd_keyring(conn, cluster, key):
    """
    Run on osd node, writes the bootstrap key if not there yet.
    """
    logger = conn.logger
    path = '/var/lib/ceph/bootstrap-osd/{cluster}.keyring'.format(
        cluster=cluster,
        )
    if not conn.remote_module.path_exists(path):
        logger.warning('osd keyring does not exist yet, creating one')
        conn.remote_module.write_keyring(path, key)


def osd_tree(conn, cluster):
    """
    Check the status of an OSD. Make sure all are up and in

    What good output would look like::

        {
            "epoch": 8,
            "num_osds": 1,
            "num_up_osds": 1,
            "num_in_osds": "1",
            "full": "false",
            "nearfull": "false"
        }

    Note how the booleans are actually strings, so we need to take that into
    account and fix it before returning the dictionary. Issue #8108
    """
    ceph_executable = system.executable_path(conn, 'ceph')
    command = [
        ceph_executable,
        '--cluster={cluster}'.format(cluster=cluster),
        'osd',
        'tree',
        '--format=json',
    ]

    out, err, code = remoto.process.check(
        conn,
        command,
    )

    try:
        loaded_json = json.loads(b''.join(out).decode('utf-8'))
        # convert boolean strings to actual booleans because
        # --format=json fails to do this properly
        for k, v in loaded_json.items():
            if v == 'true':
                loaded_json[k] = True
            elif v == 'false':
                loaded_json[k] = False
        return loaded_json
    except ValueError:
        return {}


def osd_status_check(conn, cluster):
    """
    Check the status of an OSD. Make sure all are up and in

    What good output would look like::

        {
            "epoch": 8,
            "num_osds": 1,
            "num_up_osds": 1,
            "num_in_osds": "1",
            "full": "false",
            "nearfull": "false"
        }

    Note how the booleans are actually strings, so we need to take that into
    account and fix it before returning the dictionary. Issue #8108
    """
    ceph_executable = system.executable_path(conn, 'ceph')
    command = [
        ceph_executable,
        '--cluster={cluster}'.format(cluster=cluster),
        'osd',
        'stat',
        '--format=json',
    ]

    try:
        out, err, code = remoto.process.check(
            conn,
            command,
        )
    except TypeError:
        # XXX This is a bug in remoto. If the other end disconnects with a timeout
        # it will return a None, and here we are expecting a 3 item tuple, not a None
        # so it will break with a TypeError. Once remoto fixes this, we no longer need
        # this try/except.
        return {}

    try:
        loaded_json = json.loads(b''.join(out).decode('utf-8'))
        # convert boolean strings to actual booleans because
        # --format=json fails to do this properly
        for k, v in loaded_json.items():
            if v == 'true':
                loaded_json[k] = True
            elif v == 'false':
                loaded_json[k] = False
        return loaded_json
    except ValueError:
        return {}


def catch_osd_errors(conn, logger, args):
    """
    Look for possible issues when checking the status of an OSD and
    report them back to the user.
    """
    logger.info('checking OSD status...')
    status = osd_status_check(conn, args.cluster)
    osds = int(status.get('num_osds', 0))
    up_osds = int(status.get('num_up_osds', 0))
    in_osds = int(status.get('num_in_osds', 0))
    full = status.get('full', False)
    nearfull = status.get('nearfull', False)

    if osds > up_osds:
        difference = osds - up_osds
        logger.warning('there %s %d OSD%s down' % (
            ['is', 'are'][difference != 1],
            difference,
            "s"[difference == 1:])
        )

    if osds > in_osds:
        difference = osds - in_osds
        logger.warning('there %s %d OSD%s out' % (
            ['is', 'are'][difference != 1],
            difference,
            "s"[difference == 1:])
        )

    if full:
        logger.warning('OSDs are full!')

    if nearfull:
        logger.warning('OSDs are near full!')


def prepare_disk(
        conn,
        cluster,
        disk,
        journal,
        activate_prepared_disk,
        init,
        zap,
        fs_type,
        dmcrypt,
        dmcrypt_dir,
        storetype,
        block_wal,
        block_db):
    """
    Run on osd node, prepares a data disk for use.
    """
    ceph_disk_executable = system.executable_path(conn, 'ceph-disk')
    args = [
        ceph_disk_executable,
        '-v',
        'prepare',
        ]
    if zap:
        args.append('--zap-disk')
    if dmcrypt:
        args.append('--dmcrypt')
        if dmcrypt_dir is not None:
            args.append('--dmcrypt-key-dir')
            args.append(dmcrypt_dir)
    if storetype == 'bluestore':
        if block_wal:
            args.append('--block.wal')
            args.append(block_wal)
        if block_db:
            args.append('--block.db')
            args.append(block_db)
    if storetype:
        args.append('--' + storetype)
    args.extend([
        '--cluster',
        cluster,
        '--fs-type',
        fs_type,
        '--',
        disk,
    ])

    if journal is not None:
        args.append(journal)

    remoto.process.run(
        conn,
        args
    )

    if activate_prepared_disk:
        # we don't simply run activate here because we don't know
        # which partition ceph-disk prepare created as the data
        # volume.  instead, we rely on udev to do the activation and
        # just give it a kick to ensure it wakes up.  we also enable
        # ceph.target, the other key piece of activate.
        if init == 'systemd':
            system.enable_service(conn, "ceph.target")
        elif init == 'sysvinit':
            system.enable_service(conn, "ceph")


def exceeds_max_osds(args, reasonable=20):
    """
    A very simple function to check against multiple OSDs getting created and
    warn about the possibility of more than the recommended which would cause
    issues with max allowed PIDs in a system.

    The check is done against the ``args.disk`` object that should look like::

        [
            ('cephnode-01', '/dev/sdb', '/dev/sda5'),
            ('cephnode-01', '/dev/sdc', '/dev/sda6'),
            ...
        ]
    """
    hosts = [item[0] for item in args.disk]
    per_host_count = dict(
        (
            (h, hosts.count(h)) for h in set(hosts)
            if hosts.count(h) > reasonable
        )
    )

    return per_host_count


def prepare(args, cfg, activate_prepared_disk):
    LOG.debug(
        'Preparing cluster %s disks %s',
        args.cluster,
        ' '.join(':'.join(x or '' for x in t) for t in args.disk),
        )

    hosts_in_danger = exceeds_max_osds(args)

    if hosts_in_danger:
        LOG.warning('if ``kernel.pid_max`` is not increased to a high enough value')
        LOG.warning('the following hosts will encounter issues:')
        for host, count in hosts_in_danger.items():
            LOG.warning('Host: %8s, OSDs: %s' % (host, count))

    key = get_bootstrap_osd_key(cluster=args.cluster)

    bootstrapped = set()
    errors = 0
    for hostname, disk, journal in args.disk:
        try:
            if disk is None:
                raise exc.NeedDiskError(hostname)

            distro = hosts.get(
                hostname,
                username=args.username,
                callbacks=[packages.ceph_is_installed]
            )
            LOG.info(
                'Distro info: %s %s %s',
                distro.name,
                distro.release,
                distro.codename
            )

            if hostname not in bootstrapped:
                bootstrapped.add(hostname)
                LOG.debug('Deploying osd to %s', hostname)

                conf_data = conf.ceph.load_raw(args)
                distro.conn.remote_module.write_conf(
                    args.cluster,
                    conf_data,
                    args.overwrite_conf
                )

                create_osd_keyring(distro.conn, args.cluster, key)

            LOG.debug('Preparing host %s disk %s journal %s activate %s',
                      hostname, disk, journal, activate_prepared_disk)

            storetype = None
            if args.bluestore:
                storetype = 'bluestore'

            prepare_disk(
                distro.conn,
                cluster=args.cluster,
                disk=disk,
                journal=journal,
                activate_prepared_disk=activate_prepared_disk,
                init=distro.init,
                zap=args.zap_disk,
                fs_type=args.fs_type,
                dmcrypt=args.dmcrypt,
                dmcrypt_dir=args.dmcrypt_key_dir,
                storetype=storetype,
                block_wal=args.block_wal,
                block_db=args.block_db
            )

            # give the OSD a few seconds to start
            time.sleep(5)
            catch_osd_errors(distro.conn, distro.conn.logger, args)
            LOG.debug('Host %s is now ready for osd use.', hostname)
            distro.conn.exit()

        except RuntimeError as e:
            LOG.error(e)
            errors += 1

    if errors:
        raise exc.GenericError('Failed to create %d OSDs' % errors)


def activate(args, cfg):
    LOG.debug(
        'Activating cluster %s disks %s',
        args.cluster,
        # join elements of t with ':', t's with ' '
        # allow None in elements of t; print as empty
        ' '.join(':'.join((s or '') for s in t) for t in args.disk),
        )

    for hostname, disk, journal in args.disk:

        distro = hosts.get(
            hostname,
            username=args.username,
            callbacks=[packages.ceph_is_installed]
        )
        LOG.info(
            'Distro info: %s %s %s',
            distro.name,
            distro.release,
            distro.codename
        )

        LOG.debug('activating host %s disk %s', hostname, disk)
        LOG.debug('will use init type: %s', distro.init)

        ceph_disk_executable = system.executable_path(distro.conn, 'ceph-disk')
        remoto.process.run(
            distro.conn,
            [
                ceph_disk_executable,
                '-v',
                'activate',
                '--mark-init',
                distro.init,
                '--mount',
                disk,
            ],
        )
        # give the OSD a few seconds to start
        time.sleep(5)
        catch_osd_errors(distro.conn, distro.conn.logger, args)

        if distro.init == 'systemd':
            system.enable_service(distro.conn, "ceph.target")
        elif distro.init == 'sysvinit':
            system.enable_service(distro.conn, "ceph")

        distro.conn.exit()


def disk_zap(args):

    for hostname, disk, journal in args.disk:
        if not disk or not hostname:
            raise RuntimeError('zap command needs both HOSTNAME and DISK but got "%s %s"' % (hostname, disk))
        LOG.debug('zapping %s on %s', disk, hostname)
        distro = hosts.get(
            hostname,
            username=args.username,
            callbacks=[packages.ceph_is_installed]
        )
        LOG.info(
            'Distro info: %s %s %s',
            distro.name,
            distro.release,
            distro.codename
        )

        distro.conn.remote_module.zeroing(disk)

        ceph_disk_executable = system.executable_path(distro.conn, 'ceph-disk')
        remoto.process.run(
            distro.conn,
            [
                ceph_disk_executable,
                'zap',
                disk,
            ],
        )

        distro.conn.exit()


def disk_list(args, cfg):
    for hostname, disk, journal in args.disk:
        distro = hosts.get(
            hostname,
            username=args.username,
            callbacks=[packages.ceph_is_installed]
        )
        LOG.info(
            'Distro info: %s %s %s',
            distro.name,
            distro.release,
            distro.codename
        )

        LOG.debug('Listing disks on {hostname}...'.format(hostname=hostname))
        ceph_disk_executable = system.executable_path(distro.conn, 'ceph-disk')
        remoto.process.run(
            distro.conn,
            [
                ceph_disk_executable,
                'list',
            ],
        )
        distro.conn.exit()


def osd_list(args, cfg):
    monitors = mon.get_mon_initial_members(args, error_on_empty=True, _cfg=cfg)

    # get the osd tree from a monitor host
    mon_host = monitors[0]
    distro = hosts.get(
        mon_host,
        username=args.username,
        callbacks=[packages.ceph_is_installed]
    )

    tree = osd_tree(distro.conn, args.cluster)
    distro.conn.exit()

    interesting_files = ['active', 'magic', 'whoami', 'journal_uuid']

    for hostname, disk, journal in args.disk:
        distro = hosts.get(hostname, username=args.username)
        remote_module = distro.conn.remote_module
        osds = distro.conn.remote_module.listdir(constants.osd_path)

        ceph_disk_executable = system.executable_path(distro.conn, 'ceph-disk')
        output, err, exit_code = remoto.process.check(
            distro.conn,
            [
                ceph_disk_executable,
                'list',
            ]
        )

        for _osd in osds:
            osd_path = os.path.join(constants.osd_path, _osd)
            journal_path = os.path.join(osd_path, 'journal')
            _id = int(_osd.split('-')[-1])  # split on dash, get the id
            osd_name = 'osd.%s' % _id
            metadata = {}
            json_blob = {}

            # piggy back from ceph-disk and get the mount point
            device = get_osd_mount_point(output, osd_name)
            if device:
                metadata['device'] = device

            # read interesting metadata from files
            for f in interesting_files:
                osd_f_path = os.path.join(osd_path, f)
                if remote_module.path_exists(osd_f_path):
                    metadata[f] = remote_module.readline(osd_f_path)

            # do we have a journal path?
            if remote_module.path_exists(journal_path):
                metadata['journal path'] = remote_module.get_realpath(journal_path)

            # is this OSD in osd tree?
            for blob in tree['nodes']:
                if blob.get('id') == _id:  # matches our OSD
                    json_blob = blob

            print_osd(
                distro.conn.logger,
                hostname,
                osd_path,
                json_blob,
                metadata,
            )

        distro.conn.exit()


def get_osd_mount_point(output, osd_name):
    """
    piggy back from `ceph-disk list` output and get the mount point
    by matching the line where the partition mentions the OSD name

    For example, if the name of the osd is `osd.1` and the output from
    `ceph-disk list` looks like this::

        /dev/sda :
         /dev/sda1 other, ext2, mounted on /boot
         /dev/sda2 other
         /dev/sda5 other, LVM2_member
        /dev/sdb :
         /dev/sdb1 ceph data, active, cluster ceph, osd.1, journal /dev/sdb2
         /dev/sdb2 ceph journal, for /dev/sdb1
        /dev/sr0 other, unknown
        /dev/sr1 other, unknown

    Then `/dev/sdb1` would be the right mount point. We piggy back like this
    because ceph-disk does *a lot* to properly calculate those values and we
    don't want to re-implement all the helpers for this.

    :param output: A list of lines from stdout
    :param osd_name: The actual osd name, like `osd.1`
    """
    for line in output:
        line_parts = re.split(r'[,\s]+', line)
        for part in line_parts:
            mount_point = line_parts[1]
            if osd_name == part:
                return mount_point


def print_osd(logger, hostname, osd_path, json_blob, metadata, journal=None):
    """
    A helper to print OSD metadata
    """
    logger.info('-'*40)
    logger.info('%s' % osd_path.split('/')[-1])
    logger.info('-'*40)
    logger.info('%-14s %s' % ('Path', osd_path))
    logger.info('%-14s %s' % ('ID', json_blob.get('id')))
    logger.info('%-14s %s' % ('Name', json_blob.get('name')))
    logger.info('%-14s %s' % ('Status', json_blob.get('status')))
    logger.info('%-14s %s' % ('Reweight', json_blob.get('reweight')))
    if journal:
        logger.info('Journal: %s' % journal)
    for k, v in metadata.items():
        logger.info("%-13s  %s" % (k.capitalize(), v))

    logger.info('-'*40)


def osd(args):
    cfg = conf.ceph.load(args)

    if args.subcommand == 'list':
        osd_list(args, cfg)
    elif args.subcommand == 'prepare':
        prepare(args, cfg, activate_prepared_disk=False)
    elif args.subcommand == 'create':
        prepare(args, cfg, activate_prepared_disk=True)
    elif args.subcommand == 'activate':
        activate(args, cfg)
    else:
        LOG.error('subcommand %s not implemented', args.subcommand)
        sys.exit(1)


def disk(args):
    cfg = conf.ceph.load(args)

    if args.subcommand == 'list':
        disk_list(args, cfg)
    elif args.subcommand == 'prepare':
        prepare(args, cfg, activate_prepared_disk=False)
    elif args.subcommand == 'activate':
        activate(args, cfg)
    elif args.subcommand == 'zap':
        disk_zap(args)
    else:
        LOG.error('subcommand %s not implemented', args.subcommand)
        sys.exit(1)


def colon_separated(s):
    journal = None
    disk = None
    host = None
    if s.count(':') == 2:
        (host, disk, journal) = s.split(':')
    elif s.count(':') == 1:
        (host, disk) = s.split(':')
    elif s.count(':') == 0:
        (host) = s
    else:
        raise argparse.ArgumentTypeError('must be in form HOST:DISK[:JOURNAL]')

    if disk:
        # allow just "sdb" to mean /dev/sdb
        disk = os.path.join('/dev', disk)
        if journal is not None:
            journal = os.path.join('/dev', journal)

    return (host, disk, journal)


@priority(50)
def make(parser):
    """
    Prepare a data disk on remote host.
    """
    sub_command_help = dedent("""
    Manage OSDs by preparing a data disk on remote host.

    For paths, first prepare and then activate:

        ceph-deploy osd prepare {osd-node-name}:/path/to/osd
        ceph-deploy osd activate {osd-node-name}:/path/to/osd

    For disks or journals the `create` command will do prepare and activate
    for you.
    """
    )
    parser.formatter_class = argparse.RawDescriptionHelpFormatter
    parser.description = sub_command_help

    osd_parser = parser.add_subparsers(dest='subcommand')
    osd_parser.required = True

    osd_list = osd_parser.add_parser(
        'list',
        help='List OSD info from remote host(s)'
        )
    osd_list.add_argument(
        'disk',
        nargs='+',
        metavar='HOST:DISK[:JOURNAL]',
        type=colon_separated,
        help='remote host to list OSDs from'
        )

    osd_create = osd_parser.add_parser(
        'create',
        help='Create new Ceph OSD daemon by preparing and activating disk'
        )
    osd_create.add_argument(
        '--zap-disk',
        action='store_true',
        help='destroy existing partition table and content for DISK',
        )
    osd_create.add_argument(
        '--fs-type',
        metavar='FS_TYPE',
        choices=['xfs',
                 'btrfs'
                 ],
        default='xfs',
        help='filesystem to use to format DISK (xfs, btrfs)',
        )
    osd_create.add_argument(
        '--dmcrypt',
        action='store_true',
        help='use dm-crypt on DISK',
        )
    osd_create.add_argument(
        '--dmcrypt-key-dir',
        metavar='KEYDIR',
        default='/etc/ceph/dmcrypt-keys',
        help='directory where dm-crypt keys are stored',
        )
    osd_create.add_argument(
        '--bluestore',
        action='store_true', default=None,
        help='bluestore objectstore',
        )
    osd_create.add_argument(
        '--block-db',
        default=None,
        help='bluestore block.db path'
        )
    osd_create.add_argument(
        '--block-wal',
        default=None,
        help='bluestore block.wal path'
        )
    osd_create.add_argument(
        'disk',
        nargs='+',
        metavar='HOST:DISK[:JOURNAL]',
        type=colon_separated,
        help='host and disk to prepare',
        )

    osd_prepare = osd_parser.add_parser(
        'prepare',
        help='Prepare a disk for use as Ceph OSD by formatting/partitioning disk'
        )
    osd_prepare.add_argument(
        '--zap-disk',
        action='store_true',
        help='destroy existing partition table and content for DISK',
        )
    osd_prepare.add_argument(
        '--fs-type',
        metavar='FS_TYPE',
        choices=['xfs',
                 'btrfs'
                 ],
        default='xfs',
        help='filesystem to use to format DISK (xfs, btrfs)',
        )
    osd_prepare.add_argument(
        '--dmcrypt',
        action='store_true',
        help='use dm-crypt on DISK',
        )
    osd_prepare.add_argument(
        '--dmcrypt-key-dir',
        metavar='KEYDIR',
        default='/etc/ceph/dmcrypt-keys',
        help='directory where dm-crypt keys are stored',
        )
    osd_prepare.add_argument(
        '--bluestore',
        action='store_true', default=None,
        help='bluestore objectstore',
        )
    osd_prepare.add_argument(
        '--block-db',
        default=None,
        help='bluestore block.db path'
        )
    osd_prepare.add_argument(
        '--block-wal',
        default=None,
        help='bluestore block.wal path'
        )
    osd_prepare.add_argument(
        'disk',
        nargs='+',
        metavar='HOST:DISK[:JOURNAL]',
        type=colon_separated,
        help='host and disk to prepare',
        )

    osd_activate = osd_parser.add_parser(
        'activate',
        help='Start (activate) Ceph OSD from disk that was previously prepared'
        )
    osd_activate.add_argument(
        'disk',
        nargs='+',
        metavar='HOST:DISK[:JOURNAL]',
        type=colon_separated,
        help='host and disk to activate',
        )
    parser.set_defaults(
        func=osd,
        )


@priority(50)
def make_disk(parser):
    """
    Manage disks on a remote host.
    """
    disk_parser = parser.add_subparsers(dest='subcommand')
    disk_parser.required = True

    disk_zap = disk_parser.add_parser(
        'zap',
        help='destroy existing partition table and content for DISK',
        )
    disk_zap.add_argument(
        'disk',
        nargs='+',
        metavar='HOST:DISK',
        type=colon_separated,
        help='host and disk'
        )

    disk_list = disk_parser.add_parser(
        'list',
        help='List disk info from remote host(s)'
        )
    disk_list.add_argument(
        'disk',
        nargs='+',
        metavar='HOST:DISK',
        type=colon_separated,
        help='remote host to list OSDs from'
        )

    disk_prepare = disk_parser.add_parser(
        'prepare',
        help='Prepare a disk for use as Ceph OSD by formatting/partitioning disk'
        )
    disk_prepare.add_argument(
        '--zap-disk',
        action='store_true',
        help='destroy existing partition table and content for DISK',
        )
    disk_prepare.add_argument(
        '--fs-type',
        metavar='FS_TYPE',
        choices=['xfs',
                 'btrfs'
                 ],
        default='xfs',
        help='filesystem to use to format DISK (xfs, btrfs)',
        )
    disk_prepare.add_argument(
        '--dmcrypt',
        action='store_true',
        help='use dm-crypt on DISK',
        )
    disk_prepare.add_argument(
        '--dmcrypt-key-dir',
        metavar='KEYDIR',
        default='/etc/ceph/dmcrypt-keys',
        help='directory where dm-crypt keys are stored',
        )
    disk_prepare.add_argument(
        '--bluestore',
        action='store_true', default=None,
        help='bluestore objectstore',
        )
    disk_prepare.add_argument(
        '--block-db',
        default=None,
        help='bluestore block.db path'
        )
    disk_prepare.add_argument(
        '--block-wal',
        default=None,
        help='bluestore block.wal path'
        )
    disk_prepare.add_argument(
        'disk',
        nargs='+',
        metavar='HOST:DISK',
        type=colon_separated,
        help='host and disk to prepare',
        )

    disk_activate = disk_parser.add_parser(
        'activate',
        help='Start (activate) Ceph OSD from disk that was previously prepared'
        )
    disk_activate.add_argument(
        'disk',
        nargs='+',
        metavar='HOST:DISK',
        type=colon_separated,
        help='host and disk to activate',
        )
    parser.set_defaults(
        func=disk,
        )



from ceph_deploy import gatherkeys
from ceph_deploy import new
import mock
import tempfile
import shutil
import os
import pytest


class mock_conn(object):
    def __init__(self):
        pass

class mock_distro(object):
    def __init__(self):
        self.conn = mock_conn()

class mock_rlogger(object):
    def error(self, *arg):
        return

    def debug(self, *arg):
        return


def mock_remoto_process_check_success(conn, args):
    secret = new.generate_auth_key()
    out = '[mon.]\nkey = %s\ncaps mon = allow *\n' % secret
    return out.encode('utf-8').split(b'\n'), [], 0


def mock_remoto_process_check_rc_error(conn, args):
    return [b""], [b"this failed\n"], 1


class TestGatherKeysMissing(object):
    """
    Since we are testing things that effect the content a directory we should
    test in a clean empty directory.
    """

    def setup(self):
        """
        Make temp directory for tests.
        """
        self.args = mock.Mock()
        self.distro = mock_distro()
        self.test_dir = tempfile.mkdtemp()
        self.rlogger = mock_rlogger()
        self.keypath_remote = "some_path"

    def teardown(self):
        """
        Remove temp directory and content
        """
        shutil.rmtree(self.test_dir)

    @mock.patch('ceph_deploy.lib.remoto.process.check', mock_remoto_process_check_success)
    def test_success_admin(self):
        keytype = 'admin'
        rc = gatherkeys.gatherkeys_missing(
            self.args,
            self.distro,
            self.rlogger,
            self.keypath_remote,
            keytype,
            self.test_dir
            )
        assert rc is True
        keyname = gatherkeys.keytype_path_to(self.args, keytype)
        keypath_gen = os.path.join(self.test_dir, keyname)
        assert os.path.isfile(keypath_gen)

    @mock.patch('ceph_deploy.lib.remoto.process.check', mock_remoto_process_check_success)
    def test_success_mds(self):
        keytype = 'mds'
        rc = gatherkeys.gatherkeys_missing(
            self.args,
            self.distro,
            self.rlogger,
            self.keypath_remote,
            keytype,
            self.test_dir
            )
        assert rc is True
        keyname = gatherkeys.keytype_path_to(self.args, keytype)
        keypath_gen = os.path.join(self.test_dir, keyname)
        assert os.path.isfile(keypath_gen)

    @mock.patch('ceph_deploy.lib.remoto.process.check', mock_remoto_process_check_success)
    def test_success_mgr(self):
        keytype = 'mgr'
        rc = gatherkeys.gatherkeys_missing(
            self.args,
            self.distro,
            self.rlogger,
            self.keypath_remote,
            keytype,
            self.test_dir
            )
        assert rc is True
        keyname = gatherkeys.keytype_path_to(self.args, keytype)
        keypath_gen = os.path.join(self.test_dir, keyname)
        assert os.path.isfile(keypath_gen)
        
    @mock.patch('ceph_deploy.lib.remoto.process.check', mock_remoto_process_check_success)
    def test_success_osd(self):
        keytype = 'osd'
        rc = gatherkeys.gatherkeys_missing(
            self.args,
            self.distro,
            self.rlogger,
            self.keypath_remote,
            keytype,
            self.test_dir
            )
        assert rc is True
        keyname = gatherkeys.keytype_path_to(self.args, keytype)
        keypath_gen = os.path.join(self.test_dir, keyname)
        assert os.path.isfile(keypath_gen)

    @mock.patch('ceph_deploy.lib.remoto.process.check', mock_remoto_process_check_success)
    def test_success_rgw(self):
        keytype = 'rgw'
        rc = gatherkeys.gatherkeys_missing(
            self.args,
            self.distro,
            self.rlogger,
            self.keypath_remote,
            keytype,
            self.test_dir
            )
        assert rc is True
        keyname = gatherkeys.keytype_path_to(self.args, keytype)
        keypath_gen = os.path.join(self.test_dir, keyname)
        assert os.path.isfile(keypath_gen)

    @mock.patch('ceph_deploy.lib.remoto.process.check', mock_remoto_process_check_rc_error)
    def test_remoto_process_check_rc_error(self):
        keytype = 'admin'
        rc = gatherkeys.gatherkeys_missing(
            self.args,
            self.distro,
            self.rlogger,
            self.keypath_remote,
            keytype,
            self.test_dir
            )
        assert rc is False
        keyname = gatherkeys.keytype_path_to(self.args, keytype)
        keypath_gen = os.path.join(self.test_dir, keyname)
        assert not os.path.isfile(keypath_gen)

    @mock.patch('ceph_deploy.lib.remoto.process.check', mock_remoto_process_check_success)
    def test_fail_identity_missing(self):
        keytype = 'silly'
        with pytest.raises(RuntimeError):
            gatherkeys.gatherkeys_missing(
                self.args,
                self.distro,
                self.rlogger,
                self.keypath_remote,
                keytype,
                self.test_dir
                )

    @mock.patch('ceph_deploy.lib.remoto.process.check', mock_remoto_process_check_success)
    def test_fail_capabilities_missing(self):
        keytype = 'mon'
        with pytest.raises(RuntimeError):
            gatherkeys.gatherkeys_missing(
                self.args,
                self.distro,
                self.rlogger,
                self.keypath_remote,
                keytype,
                self.test_dir
                )




import errno
import os.path
import logging
import json
import tempfile
import shutil
import time

from ceph_deploy import hosts
from ceph_deploy.cliutil import priority
from ceph_deploy.lib import remoto
import ceph_deploy.util.paths.mon

LOG = logging.getLogger(__name__)


def _keyring_equivalent(keyring_one, keyring_two):
    """
    Check two keyrings are identical
    """
    def keyring_extract_key(file_path):
        """
        Cephx keyring files may or may not have white space before some lines.
        They may have some values in quotes, so a safe way to compare is to
        extract the key.
        """
        with open(file_path) as f:
            for line in f:
                content = line.strip()
                if len(content) == 0:
                    continue
                split_line = content.split('=')
                if split_line[0].strip() == 'key':
                    return "=".join(split_line[1:]).strip()
        raise RuntimeError("File '%s' is not a keyring" % file_path)
    key_one = keyring_extract_key(keyring_one)
    key_two = keyring_extract_key(keyring_two)
    return key_one == key_two


def keytype_path_to(args, keytype):
    """
    Get the local filename for a keyring type
    """
    if keytype == "admin":
        return '{cluster}.client.admin.keyring'.format(
            cluster=args.cluster)
    if keytype == "mon":
        return '{cluster}.mon.keyring'.format(
            cluster=args.cluster)
    return '{cluster}.bootstrap-{what}.keyring'.format(
            cluster=args.cluster,
            what=keytype)


def keytype_identity(keytype):
    """
    Get the keyring identity from keyring type.

    This is used in authentication with keyrings and generating keyrings.
    """
    ident_dict = {
        'admin' : 'client.admin',
        'mds' : 'client.bootstrap-mds',
        'mgr' : 'client.bootstrap-mgr',
        'osd' : 'client.bootstrap-osd',
        'rgw' : 'client.bootstrap-rgw',
        'mon' : 'mon.'
    }
    return ident_dict.get(keytype, None)


def keytype_capabilities(keytype):
    """
    Get the capabilities of a keyring from keyring type.
    """
    cap_dict = {
        'admin' : [
            'osd', 'allow *',
            'mds', 'allow *',
            'mon', 'allow *',
            'mgr', 'allow *'
            ],
        'mds' : [
            'mon', 'allow profile bootstrap-mds'
            ],
        'mgr' : [
            'mon', 'allow profile bootstrap-mgr'
            ],
        'osd' : [
            'mon', 'allow profile bootstrap-osd'
            ],
        'rgw': [
            'mon', 'allow profile bootstrap-rgw'
            ]
        }
    return cap_dict.get(keytype, None)


def gatherkeys_missing(args, distro, rlogger, keypath, keytype, dest_dir):
    """
    Get or create the keyring from the mon using the mon keyring by keytype and
    copy to dest_dir
    """
    args_prefix = [
        '/usr/bin/ceph',
        '--connect-timeout=25',
        '--cluster={cluster}'.format(
            cluster=args.cluster),
        '--name', 'mon.',
        '--keyring={keypath}'.format(
            keypath=keypath),
        ]

    identity = keytype_identity(keytype)
    if identity is None:
        raise RuntimeError('Could not find identity for keytype:%s' % keytype)
    capabilites = keytype_capabilities(keytype)
    if capabilites is None:
        raise RuntimeError('Could not find capabilites for keytype:%s' % keytype)

    # First try getting the key if it already exists, to handle the case where
    # it exists but doesn't match the caps we would pass into get-or-create.
    # This is the same behvaior as in newer ceph-create-keys
    out, err, code = remoto.process.check(
        distro.conn,
        args_prefix + ['auth', 'get', identity]
        )
    if code == errno.ENOENT:
        out, err, code = remoto.process.check(
            distro.conn,
            args_prefix + ['auth', 'get-or-create', identity] + capabilites
            )
    if code != 0:
        rlogger.error(
            '"ceph auth get-or-create for keytype %s returned %s',
            keytype, code
        )
        for line in err:
            rlogger.debug(line)
        return False
    keyring_name_local = keytype_path_to(args, keytype)
    keyring_path_local = os.path.join(dest_dir, keyring_name_local)
    with open(keyring_path_local, 'wb') as f:
        for line in out:
            f.write(line + b'\n')
    return True


def gatherkeys_with_mon(args, host, dest_dir):
    """
    Connect to mon and gather keys if mon is in quorum.
    """
    distro = hosts.get(host, username=args.username)
    remote_hostname = distro.conn.remote_module.shortname()
    dir_keytype_mon = ceph_deploy.util.paths.mon.path(args.cluster, remote_hostname)
    path_keytype_mon = "%s/keyring" % (dir_keytype_mon)
    mon_key = distro.conn.remote_module.get_file(path_keytype_mon)
    if mon_key is None:
        LOG.warning("No mon key found in host: %s", host)
        return False
    mon_name_local = keytype_path_to(args, "mon")
    mon_path_local = os.path.join(dest_dir, mon_name_local)
    with open(mon_path_local, 'wb') as f:
        f.write(mon_key)
    rlogger = logging.getLogger(host)
    path_asok = ceph_deploy.util.paths.mon.asok(args.cluster, remote_hostname)
    out, err, code = remoto.process.check(
        distro.conn,
            [
                "/usr/bin/ceph",
                "--connect-timeout=25",
                "--cluster={cluster}".format(
                    cluster=args.cluster),
                "--admin-daemon={asok}".format(
                    asok=path_asok),
                "mon_status"
            ]
        )
    if code != 0:
        rlogger.error('"ceph mon_status %s" returned %s', host, code)
        for line in err:
            rlogger.debug(line)
        return False
    try:
        mon_status = json.loads(b''.join(out).decode('utf-8'))
    except ValueError:
        rlogger.error('"ceph mon_status %s" output was not json', host)
        for line in out:
            rlogger.error(line)
        return False
    mon_number = None
    mon_map = mon_status.get('monmap')
    if mon_map is None:
        rlogger.error("could not find mon map for mons on '%s'", host)
        return False
    mon_quorum = mon_status.get('quorum')
    if mon_quorum is None:
        rlogger.error("could not find quorum for mons on '%s'" , host)
        return False
    mon_map_mons = mon_map.get('mons')
    if mon_map_mons is None:
        rlogger.error("could not find mons in monmap on '%s'", host)
        return False
    for mon in mon_map_mons:
        if mon.get('name') == remote_hostname:
           mon_number = mon.get('rank')
           break
    if mon_number is None:
        rlogger.error("could not find '%s' in monmap", remote_hostname)
        return False
    if not mon_number in mon_quorum:
        rlogger.error("Not yet quorum for '%s'", host)
        return False
    for keytype in ["admin", "mds", "mgr", "osd", "rgw"]:
        if not gatherkeys_missing(args, distro, rlogger, path_keytype_mon, keytype, dest_dir):
            # We will return failure if we fail to gather any key
            rlogger.error("Failed to return '%s' key from host %s", keytype, host)
            return False
    return True


def gatherkeys(args):
    """
    Gather keys from any mon and store in current working directory.

    Backs up keys from previous installs and stores new keys.
    """
    oldmask = os.umask(0o77)
    try:
        try:
            tmpd = tempfile.mkdtemp()
            LOG.info("Storing keys in temp directory %s", tmpd)
            sucess = False
            for host in args.mon:
                sucess = gatherkeys_with_mon(args, host, tmpd)
                if sucess:
                    break
            if not sucess:
                LOG.error("Failed to connect to host:%s" ,', '.join(args.mon))
                raise RuntimeError('Failed to connect any mon')
            had_error = False
            date_string = time.strftime("%Y%m%d%H%M%S")
            for keytype in ["admin", "mds", "mgr", "mon", "osd", "rgw"]:
                filename = keytype_path_to(args, keytype)
                tmp_path = os.path.join(tmpd, filename)
                if not os.path.exists(tmp_path):
                    LOG.error("No key retrived for '%s'" , keytype)
                    had_error = True
                    continue
                if not os.path.exists(filename):
                    LOG.info("Storing %s" % (filename))
                    shutil.move(tmp_path, filename)
                    continue
                if _keyring_equivalent(tmp_path, filename):
                    LOG.info("keyring '%s' already exists" , filename)
                    continue
                backup_keyring = "%s-%s" % (filename, date_string)
                LOG.info("Replacing '%s' and backing up old key as '%s'", filename, backup_keyring)
                shutil.copy(filename, backup_keyring)
                shutil.move(tmp_path, filename)
            if had_error:
                raise RuntimeError('Failed to get all key types')
        finally:
            LOG.info("Destroy temp directory %s" %(tmpd))
            shutil.rmtree(tmpd)
    finally:
        os.umask(oldmask)


@priority(40)
def make(parser):
    """
    Gather authentication keys for provisioning new nodes.
    """
    parser.add_argument(
        'mon',
        metavar='HOST',
        nargs='+',
        help='monitor host to pull keys from',
        )
    parser.set_defaults(
        func=gatherkeys,
        )



try:
    from urllib.parse import urlparse
except ImportError:
    from urlparse import urlparse
import logging
from ceph_deploy.util.paths import gpg
from ceph_deploy.util import net


LOG = logging.getLogger(__name__)


def install(distro, version_kind, version, adjust_repos, **kw):
    packages = kw.pop('components', [])
    codename = distro.codename
    machine = distro.machine_type
    extra_install_flags = []

    if version_kind in ['stable', 'testing']:
        key = 'release'
    else:
        key = 'autobuild'

    distro.packager.install(['ca-certificates', 'apt-transport-https'])

    if adjust_repos:
        # Wheezy does not like the download.ceph.com SSL cert
        protocol = 'https'
        if codename == 'wheezy':
            protocol = 'http'

        if version_kind in ['dev', 'dev_commit']:
            shaman_url = 'https://shaman.ceph.com/api/repos/ceph/{version}/{sha1}/{distro}/{distro_version}/repo/?arch={arch}'.format(
                distro=distro.normalized_name,
                distro_version=distro.codename,
                version=kw['args'].dev,
                sha1=kw['args'].dev_commit or 'latest',
                arch=machine
                )
            LOG.debug('fetching repo information from: %s' % shaman_url)
            chacra_url = net.get_request(shaman_url).geturl()
            content = net.get_chacra_repo(shaman_url)
            # set the repo priority for the right domain
            fqdn = urlparse(chacra_url).hostname
            distro.conn.remote_module.set_apt_priority(fqdn)
            distro.conn.remote_module.write_sources_list_content(content)
            extra_install_flags = ['-o', 'Dpkg::Options::=--force-confnew', '--allow-unauthenticated']
        else:
            distro.packager.add_repo_gpg_key(gpg.url(key, protocol=protocol))
            if version_kind == 'stable':
                url = '{protocol}://download.ceph.com/debian-{version}/'.format(
                    protocol=protocol,
                    version=version,
                    )
            elif version_kind == 'testing':
                url = '{protocol}://download.ceph.com/debian-testing/'.format(
                    protocol=protocol,
                    )
            else:
                raise RuntimeError('Unknown version kind: %r' % version_kind)

            # set the repo priority for the right domain
            fqdn = urlparse(url).hostname
            distro.conn.remote_module.set_apt_priority(fqdn)
            distro.conn.remote_module.write_sources_list(url, codename)
            extra_install_flags = ['-o', 'Dpkg::Options::=--force-confnew']

    distro.packager.clean()

    # TODO this does not downgrade -- should it?
    if packages:
        distro.packager.install(
            packages,
            extra_install_flags=extra_install_flags
        )


def mirror_install(distro, repo_url, gpg_url, adjust_repos, **kw):
    packages = kw.pop('components', [])
    version_kind = kw['args'].version_kind
    repo_url = repo_url.strip('/')  # Remove trailing slashes

    if adjust_repos:
        distro.packager.add_repo_gpg_key(gpg_url)

        # set the repo priority for the right domain
        fqdn = urlparse(repo_url).hostname
        distro.conn.remote_module.set_apt_priority(fqdn)

        distro.conn.remote_module.write_sources_list(repo_url, distro.codename)

    extra_install_flags = ['--allow-unauthenticated'] if version_kind in 'dev' else []

    if packages:
        distro.packager.clean()
        distro.packager.install(
            packages,
            extra_install_flags=extra_install_flags)


def repo_install(distro, repo_name, baseurl, gpgkey, **kw):
    packages = kw.pop('components', [])
    # Get some defaults
    safe_filename = '%s.list' % repo_name.replace(' ', '-')
    install_ceph = kw.pop('install_ceph', False)
    baseurl = baseurl.strip('/')  # Remove trailing slashes

    distro.packager.add_repo_gpg_key(gpgkey)

    distro.conn.remote_module.write_sources_list(
        baseurl,
        distro.codename,
        safe_filename
    )

    # set the repo priority for the right domain
    fqdn = urlparse(baseurl).hostname
    distro.conn.remote_module.set_apt_priority(fqdn)

    # repo is not operable until an update
    distro.packager.clean()

    if install_ceph and packages:
        distro.packager.install(packages)





def reporting(conn, result, timeout=None):
    timeout = timeout or conn.global_timeout # -1 a.k.a. wait for ever
    log_map = {
        'debug': conn.logger.debug,
        'error': conn.logger.error,
        'warning': conn.logger.warning
    }

    while True:
        try:
            received = result.receive(timeout)
            level_received, message = list(received.items())[0]
            if not isinstance(message, str):
                message = message.decode('utf-8')
            log_map[level_received](message.strip('\n'))
        except EOFError:
            break
        except Exception as err:
            # the things we need to do here :(
            # because execnet magic, we cannot catch this as
            # `except TimeoutError`
            if err.__class__.__name__ == 'TimeoutError':
                msg = 'No data was received after %s seconds, disconnecting...' % timeout
                conn.logger.warning(msg)
                break
            raise



import pytest

from ceph_deploy.cli import get_parser
from ceph_deploy.tests.util import assert_too_few_arguments

COMP_FLAGS = [
    'mon', 'mds', 'rgw', 'osd', 'common', 'all'
]


class TestParserInstall(object):

    def setup(self):
        self.parser = get_parser()

    def test_install_help(self, capsys):
        with pytest.raises(SystemExit):
            self.parser.parse_args('install --help'.split())
        out, err = capsys.readouterr()
        assert 'usage: ceph-deploy install' in out
        assert 'positional arguments:' in out
        assert 'optional arguments:' in out

    def test_install_host_required(self, capsys):
        with pytest.raises(SystemExit):
            self.parser.parse_args('install'.split())
        out, err = capsys.readouterr()
        assert_too_few_arguments(err)

    def test_install_one_host(self):
        args = self.parser.parse_args('install host1'.split())
        assert args.host == ['host1']

    def test_install_multiple_hosts(self):
        hostnames = ['host1', 'host2', 'host3']
        args = self.parser.parse_args(['install'] + hostnames)
        assert frozenset(args.host) == frozenset(hostnames)

    def test_install_release_default_is_none(self):
        args = self.parser.parse_args('install host1'.split())
        assert args.release is None
        assert args.version_kind == "stable"

    def test_install_release(self):
        args = self.parser.parse_args('install --release hammer host1'.split())
        assert args.release == "hammer"
        assert args.version_kind == "stable"

    @pytest.mark.skipif(reason="No release name sanity checking yet")
    def test_install_release_bad_codename(self):
        args = self.parser.parse_args('install --release cephalopod host1'.split())
        assert args.release != "cephalopod"

    def test_install_testing_default_is_none(self):
        args = self.parser.parse_args('install host1'.split())
        assert args.testing is None
        assert args.version_kind == "stable"

    def test_install_testing_true(self):
        args = self.parser.parse_args('install --testing host1'.split())
        assert len(args.testing) == 0
        assert args.version_kind == "testing"

    def test_install_dev_disabled_by_default(self):
        args = self.parser.parse_args('install host1'.split())
        # dev defaults to master, but version_kind nullifies it
        assert args.dev == "master"
        assert args.version_kind == "stable"

    def test_install_dev_custom_version(self):
        args = self.parser.parse_args('install --dev v0.80.8 host1'.split())
        assert args.dev == "v0.80.8"
        assert args.version_kind == "dev"

    @pytest.mark.skipif(reason="test reflects desire, but not code reality")
    def test_install_dev_option_default_is_master(self):
        # I don't think this is the way argparse works.
        args = self.parser.parse_args('install --dev host1'.split())
        assert args.dev == "master"
        assert args.version_kind == "dev"

    def test_install_release_testing_mutex(self, capsys):
        with pytest.raises(SystemExit):
            self.parser.parse_args('install --release hammer --testing host1'.split())
        out, err = capsys.readouterr()
        assert 'not allowed with argument' in err

    def test_install_release_dev_mutex(self, capsys):
        with pytest.raises(SystemExit):
            self.parser.parse_args('install --release hammer --dev master host1'.split())
        out, err = capsys.readouterr()
        assert 'not allowed with argument' in err

    def test_install_testing_dev_mutex(self, capsys):
        with pytest.raises(SystemExit):
            self.parser.parse_args('install --testing --dev master host1'.split())
        out, err = capsys.readouterr()
        assert 'not allowed with argument' in err

    @pytest.mark.parametrize('comp', COMP_FLAGS)
    def test_install_component_default_is_false(self, comp):
        args = self.parser.parse_args('install host1'.split())
        assert getattr(args, 'install_%s' % comp) is False

    @pytest.mark.parametrize('comp', COMP_FLAGS)
    def test_install_component_true(self, comp):
        args = self.parser.parse_args(('install --%s host1' % comp).split())
        assert getattr(args, 'install_%s' % comp) is True

    def test_install_multi_component(self):
        args = self.parser.parse_args(('install --mon --rgw host1').split())
        assert args.install_mon
        assert args.install_rgw

    def test_install_adjust_repos_default_is_true(self):
        args = self.parser.parse_args('install host1'.split())
        assert args.adjust_repos

    def test_install_adjust_repos_false(self):
        args = self.parser.parse_args('install --no-adjust-repos host1'.split())
        assert not args.adjust_repos

    def test_install_adjust_repos_false_with_custom_release(self):
        args = self.parser.parse_args('install --release firefly --no-adjust-repos host1'.split())
        assert args.release == "firefly"
        assert not args.adjust_repos

    def test_install_repo_default_is_false(self):
        args = self.parser.parse_args('install host1'.split())
        assert not args.repo

    def test_install_repo_true(self):
        args = self.parser.parse_args('install --repo host1'.split())
        assert args.repo

    def test_install_local_mirror_default_is_none(self):
        args = self.parser.parse_args('install host1'.split())
        assert args.local_mirror is None

    def test_install_local_mirror_custom_path(self):
        args = self.parser.parse_args('install --local-mirror /mnt/mymirror host1'.split())
        assert args.local_mirror == "/mnt/mymirror"

    def test_install_repo_url_default_is_none(self):
        args = self.parser.parse_args('install host1'.split())
        assert args.repo_url is None

    def test_install_repo_url_custom_path(self):
        args = self.parser.parse_args('install --repo-url https://ceph.com host1'.split())
        assert args.repo_url == "https://ceph.com"

    def test_install_gpg_url_default_is_none(self):
        args = self.parser.parse_args('install host1'.split())
        assert args.gpg_url is None

    def test_install_gpg_url_custom_path(self):
        args = self.parser.parse_args('install --gpg-url https://ceph.com/key host1'.split())
        assert args.gpg_url == "https://ceph.com/key"



import errno
import logging
import os
import uuid
import struct
import time
import base64
import socket

from ceph_deploy.cliutil import priority
from ceph_deploy import conf, hosts, exc
from ceph_deploy.util import arg_validators, ssh, net
from ceph_deploy.misc import mon_hosts
from ceph_deploy.lib import remoto
from ceph_deploy.connection import get_local_connection


LOG = logging.getLogger(__name__)


def generate_auth_key():
    key = os.urandom(16)
    header = struct.pack(
        '<hiih',
        1,                 # le16 type: CEPH_CRYPTO_AES
        int(time.time()),  # le32 created: seconds
        0,                 # le32 created: nanoseconds,
        len(key),          # le16: len(key)
    )
    return base64.b64encode(header + key).decode('utf-8')


def ssh_copy_keys(hostname, username=None):
    LOG.info('making sure passwordless SSH succeeds')
    if ssh.can_connect_passwordless(hostname):
        return

    LOG.warning('could not connect via SSH')

    # Create the key if it doesn't exist:
    id_rsa_pub_file = os.path.expanduser(u'~/.ssh/id_rsa.pub')
    id_rsa_file = id_rsa_pub_file.split('.pub')[0]
    if not os.path.exists(id_rsa_file):
        LOG.info('creating a passwordless id_rsa.pub key file')
        with get_local_connection(LOG) as conn:
            remoto.process.run(
                conn,
                [
                    'ssh-keygen',
                    '-t',
                    'rsa',
                    '-N',
                    "",
                    '-f',
                    id_rsa_file,
                ]
            )

    # Get the contents of id_rsa.pub and push it to the host
    LOG.info('will connect again with password prompt')
    distro = hosts.get(hostname, username, detect_sudo=False)
    auth_keys_path = '.ssh/authorized_keys'
    if not distro.conn.remote_module.path_exists(auth_keys_path):
        distro.conn.logger.warning(
            '.ssh/authorized_keys does not exist, will skip adding keys'
        )
    else:
        LOG.info('adding public keys to authorized_keys')
        with open(os.path.expanduser('~/.ssh/id_rsa.pub'), 'r') as id_rsa:
            contents = id_rsa.read()
        distro.conn.remote_module.append_to_file(
            auth_keys_path,
            contents
        )
    distro.conn.exit()


def validate_host_ip(ips, subnets):
    """
    Make sure that a given host all subnets specified will have at least one IP
    in that range.
    """
    # Make sure we prune ``None`` arguments
    subnets = [s for s in subnets if s is not None]
    validate_one_subnet = len(subnets) == 1

    def ip_in_one_subnet(ips, subnet):
        """ ensure an ip exists in at least one subnet """
        for ip in ips:
            if net.ip_in_subnet(ip, subnet):
                return True
        return False

    for subnet in subnets:
        if ip_in_one_subnet(ips, subnet):
            if validate_one_subnet:
                return
            else:  # keep going to make sure the other subnets are ok
                continue
        else:
            msg = "subnet (%s) is not valid for any of the ips found %s" % (subnet, str(ips))
            raise RuntimeError(msg)


def get_public_network_ip(ips, public_subnet):
    """
    Given a public subnet, chose the one IP from the remote host that exists
    within the subnet range.
    """
    for ip in ips:
        if net.ip_in_subnet(ip, public_subnet):
            return ip
    msg = "IPs (%s) are not valid for any of subnet specified %s" % (str(ips), str(public_subnet))
    raise RuntimeError(msg)


def new(args):
    if args.ceph_conf:
        raise RuntimeError('will not create a Ceph conf file if attemtping to re-use with `--ceph-conf` flag')
    LOG.debug('Creating new cluster named %s', args.cluster)
    cfg = conf.ceph.CephConf()
    cfg.add_section('global')

    fsid = args.fsid or uuid.uuid4()
    cfg.set('global', 'fsid', str(fsid))

    # if networks were passed in, lets set them in the
    # global section
    if args.public_network:
        cfg.set('global', 'public network', str(args.public_network))

    if args.cluster_network:
        cfg.set('global', 'cluster network', str(args.cluster_network))

    mon_initial_members = []
    mon_host = []

    for (name, host) in mon_hosts(args.mon):
        # Try to ensure we can ssh in properly before anything else
        if args.ssh_copykey:
            ssh_copy_keys(host, args.username)

        # Now get the non-local IPs from the remote node
        distro = hosts.get(host, username=args.username)
        remote_ips = net.ip_addresses(distro.conn)

        # custom cluster names on sysvinit hosts won't work
        if distro.init == 'sysvinit' and args.cluster != 'ceph':
            LOG.error('custom cluster names are not supported on sysvinit hosts')
            raise exc.ClusterNameError(
                'host %s does not support custom cluster names' % host
            )

        distro.conn.exit()

        # Validate subnets if we received any
        if args.public_network or args.cluster_network:
            validate_host_ip(remote_ips, [args.public_network, args.cluster_network])

        # Pick the IP that matches the public cluster (if we were told to do
        # so) otherwise pick the first, non-local IP
        LOG.debug('Resolving host %s', host)
        if args.public_network:
            ip = get_public_network_ip(remote_ips, args.public_network)
        else:
            ip = net.get_nonlocal_ip(host)
        LOG.debug('Monitor %s at %s', name, ip)
        mon_initial_members.append(name)
        try:
            socket.inet_pton(socket.AF_INET6, ip)
            mon_host.append("[" + ip + "]")
            LOG.info('Monitors are IPv6, binding Messenger traffic on IPv6')
            cfg.set('global', 'ms bind ipv6', 'true')
        except socket.error:
            mon_host.append(ip)



    LOG.debug('Monitor initial members are %s', mon_initial_members)
    LOG.debug('Monitor addrs are %s', mon_host)

    cfg.set('global', 'mon initial members', ', '.join(mon_initial_members))
    # no spaces here, see http://tracker.newdream.net/issues/3145
    cfg.set('global', 'mon host', ','.join(mon_host))

    # override undesirable defaults, needed until bobtail

    # http://tracker.ceph.com/issues/6788
    cfg.set('global', 'auth cluster required', 'cephx')
    cfg.set('global', 'auth service required', 'cephx')
    cfg.set('global', 'auth client required', 'cephx')

    path = '{name}.conf'.format(
        name=args.cluster,
        )

    new_mon_keyring(args)

    LOG.debug('Writing initial config to %s...', path)
    tmp = '%s.tmp' % path
    with open(tmp, 'w') as f:
        cfg.write(f)
    try:
        os.rename(tmp, path)
    except OSError as e:
        if e.errno == errno.EEXIST:
            raise exc.ClusterExistsError(path)
        else:
            raise


def new_mon_keyring(args):
    LOG.debug('Creating a random mon key...')
    mon_keyring = '[mon.]\nkey = %s\ncaps mon = allow *\n' % generate_auth_key()

    keypath = '{name}.mon.keyring'.format(
        name=args.cluster,
        )
    oldmask = os.umask(0o77)
    LOG.debug('Writing monitor keyring to %s...', keypath)
    try:
        tmp = '%s.tmp' % keypath
        with open(tmp, 'w', 0o600) as f:
            f.write(mon_keyring)
        try:
            os.rename(tmp, keypath)
        except OSError as e:
            if e.errno == errno.EEXIST:
                raise exc.ClusterExistsError(keypath)
            else:
                raise
    finally:
        os.umask(oldmask)


@priority(10)
def make(parser):
    """
    Start deploying a new cluster, and write a CLUSTER.conf and keyring for it.
    """
    parser.add_argument(
        'mon',
        metavar='MON',
        nargs='+',
        help='initial monitor hostname, fqdn, or hostname:fqdn pair',
        type=arg_validators.Hostname(),
        )
    parser.add_argument(
        '--no-ssh-copykey',
        dest='ssh_copykey',
        action='store_false',
        default=True,
        help='do not attempt to copy SSH keys',
    )

    parser.add_argument(
        '--fsid',
        dest='fsid',
        help='provide an alternate FSID for ceph.conf generation',
    )

    parser.add_argument(
        '--cluster-network',
        help='specify the (internal) cluster network',
        type=arg_validators.Subnet(),
    )

    parser.add_argument(
        '--public-network',
        help='specify the public network for a cluster',
        type=arg_validators.Subnet(),
    )

    parser.set_defaults(
        func=new,
        )



from mock import patch
from ceph_deploy.hosts import remotes
from ceph_deploy.hosts.remotes import platform_information

class FakeExists(object):

    def __init__(self, existing_paths):
        self.existing_paths = existing_paths

    def __call__(self, path):
        for existing_path in self.existing_paths:
            if path == existing_path:
                return path


class TestWhich(object):

    def setup(self):
        self.exists_module = 'ceph_deploy.hosts.remotes.os.path.exists'

    def test_finds_absolute_paths(self):
        exists = FakeExists(['/bin/ls'])
        with patch(self.exists_module, exists):
            path = remotes.which('ls')
        assert path == '/bin/ls'

    def test_does_not_find_executable(self):
        exists = FakeExists(['/bin/foo'])
        with patch(self.exists_module, exists):
            path = remotes.which('ls')
        assert path is None

class TestPlatformInformation(object):
    """ tests various inputs that remotes.platform_information handles

    you can test your OS string by comparing the results with the output from:
      python -c "import platform; print platform.linux_distribution()"
    """

    def setup(self):
        pass

    def test_handles_deb_version_num(self):
        def fake_distro(): return ('debian', '8.4', '')
        distro, release, codename = platform_information(fake_distro)
        assert distro == 'debian'
        assert release == '8.4'
        assert codename == 'jessie'

    def test_handles_deb_version_slash(self):
        def fake_distro(): return ('debian', 'wheezy/something', '')
        distro, release, codename = platform_information(fake_distro)
        assert distro == 'debian'
        assert release == 'wheezy/something'
        assert codename == 'wheezy'

    def test_handles_deb_version_slash_sid(self):
        def fake_distro(): return ('debian', 'jessie/sid', '')
        distro, release, codename = platform_information(fake_distro)
        assert distro == 'debian'
        assert release == 'jessie/sid'
        assert codename == 'sid'

    def test_handles_no_codename(self):
        def fake_distro(): return ('SlaOS', '99.999', '')
        distro, release, codename = platform_information(fake_distro)
        assert distro == 'SlaOS'
        assert release == '99.999'
        assert codename == ''

    # Normal distro strings
    def test_hanles_centos_64(self):
        def fake_distro(): return ('CentOS', '6.4', 'Final')
        distro, release, codename = platform_information(fake_distro)
        assert distro == 'CentOS'
        assert release == '6.4'
        assert codename == 'Final'


    def test_handles_ubuntu_percise(self):
        def fake_distro(): return ('Ubuntu', '12.04', 'precise')
        distro, release, codename = platform_information(fake_distro)
        assert distro == 'Ubuntu'
        assert release == '12.04'
        assert codename == 'precise'



from ceph_deploy import gatherkeys
from ceph_deploy import new
import mock
import pytest
import tempfile
import os
import shutil


def get_key_static(keytype, key_path):
    with open(key_path, 'w') as f:
        f.write("[%s]\n" % (gatherkeys.keytype_identity(keytype)))
        f.write("key=fred\n")


def get_key_dynamic(keytype, key_path):
    with open(key_path, 'w', 0o600) as f:
        f.write("[%s]\n" % (gatherkeys.keytype_identity(keytype)))
        f.write("key='%s'" % (new.generate_auth_key()))


def mock_time_strftime(time_format):
    return "20160412144231"


def mock_get_keys_fail(args, host, dest_dir):
    return False


def mock_get_keys_sucess_static(args, host, dest_dir):
    for keytype in ["admin", "mon", "osd", "mds", "mgr", "rgw"]:
        keypath = gatherkeys.keytype_path_to(args, keytype)
        path = "%s/%s" % (dest_dir, keypath)
        get_key_static(keytype, path)
    return True


def mock_get_keys_sucess_dynamic(args, host, dest_dir):
    for keytype in ["admin", "mon", "osd", "mds", "mgr", "rgw"]:
        keypath = gatherkeys.keytype_path_to(args, keytype)
        path = "%s/%s" % (dest_dir, keypath)
        get_key_dynamic(keytype, path)
    return True


class TestGatherKeys(object):
    """
    Since we are testing things that effect the content of the current working
    directory we should test in a clean empty directory.
    """
    def setup(self):
        """
        Make temp directory for tests and set as current working directory
        """
        self.orginaldir = os.getcwd()
        self.test_dir = tempfile.mkdtemp()
        os.chdir(self.test_dir)


    def teardown(self):
        """
        Set current working directory to old value
        Remove temp directory and content
        """
        os.chdir(self.orginaldir)
        shutil.rmtree(self.test_dir)


    @mock.patch('ceph_deploy.gatherkeys.gatherkeys_with_mon', mock_get_keys_fail)
    def test_gatherkeys_fail(self):
        """
        Test 'gatherkeys' fails when connecting to mon fails.
        """
        args = mock.Mock()
        args.cluster = "ceph"
        args.mon = ['host1']
        with pytest.raises(RuntimeError):
            gatherkeys.gatherkeys(args)


    @mock.patch('ceph_deploy.gatherkeys.gatherkeys_with_mon', mock_get_keys_sucess_static)
    def test_gatherkeys_success(self):
        """
        Test 'gatherkeys' succeeds when getinig keys that are always the same.
        Test 'gatherkeys' does not backup identical keys
        """
        args = mock.Mock()
        args.cluster = "ceph"
        args.mon = ['host1']
        gatherkeys.gatherkeys(args)
        dir_content = os.listdir(self.test_dir)
        assert "ceph.client.admin.keyring" in dir_content
        assert "ceph.bootstrap-mds.keyring" in dir_content
        assert "ceph.bootstrap-mgr.keyring" in dir_content
        assert "ceph.mon.keyring" in dir_content
        assert "ceph.bootstrap-osd.keyring" in dir_content
        assert "ceph.bootstrap-rgw.keyring" in dir_content
        assert len(dir_content) == 6
        # Now we repeat as no new keys are generated
        gatherkeys.gatherkeys(args)
        dir_content = os.listdir(self.test_dir)
        assert len(dir_content) == 6


    @mock.patch('ceph_deploy.gatherkeys.time.strftime', mock_time_strftime)
    @mock.patch('ceph_deploy.gatherkeys.gatherkeys_with_mon', mock_get_keys_sucess_dynamic)
    def test_gatherkeys_backs_up(self):
        """
        Test 'gatherkeys' succeeds when getting keys that are always different.
        Test 'gatherkeys' does backup keys that are not identical.
        """
        args = mock.Mock()
        args.cluster = "ceph"
        args.mon = ['host1']
        gatherkeys.gatherkeys(args)
        dir_content = os.listdir(self.test_dir)
        assert "ceph.client.admin.keyring" in dir_content
        assert "ceph.bootstrap-mds.keyring" in dir_content
        assert "ceph.bootstrap-mgr.keyring" in dir_content
        assert "ceph.mon.keyring" in dir_content
        assert "ceph.bootstrap-osd.keyring" in dir_content
        assert "ceph.bootstrap-rgw.keyring" in dir_content
        assert len(dir_content) == 6
        # Now we repeat as new keys are generated and old
        # are backed up
        gatherkeys.gatherkeys(args)
        dir_content = os.listdir(self.test_dir)
        mocked_time = mock_time_strftime(None)
        assert "ceph.client.admin.keyring" in dir_content
        assert "ceph.bootstrap-mds.keyring" in dir_content
        assert "ceph.bootstrap-mgr.keyring" in dir_content
        assert "ceph.mon.keyring" in dir_content
        assert "ceph.bootstrap-osd.keyring" in dir_content
        assert "ceph.bootstrap-rgw.keyring" in dir_content
        assert "ceph.client.admin.keyring-%s" % (mocked_time) in dir_content
        assert "ceph.bootstrap-mds.keyring-%s" % (mocked_time) in dir_content
        assert "ceph.bootstrap-mgr.keyring-%s" % (mocked_time) in dir_content
        assert "ceph.mon.keyring-%s" % (mocked_time) in dir_content
        assert "ceph.bootstrap-osd.keyring-%s" % (mocked_time) in dir_content
        assert "ceph.bootstrap-rgw.keyring-%s" % (mocked_time) in dir_content
        assert len(dir_content) == 12



import pytest

from ceph_deploy.cli import get_parser
from ceph_deploy.tests.util import assert_too_few_arguments


class TestParserPkg(object):

    def setup(self):
        self.parser = get_parser()

    def test_pkg_help(self, capsys):
        with pytest.raises(SystemExit):
            self.parser.parse_args('pkg --help'.split())
        out, err = capsys.readouterr()
        assert 'usage: ceph-deploy pkg' in out
        assert 'positional arguments:' in out
        assert 'optional arguments:' in out

    def test_pkg_install_host_required(self, capsys):
        with pytest.raises(SystemExit):
            self.parser.parse_args('pkg --install pkg1'.split())
        out, err = capsys.readouterr()
        assert_too_few_arguments(err)

    def test_pkg_install_one_host(self):
        args = self.parser.parse_args('pkg --install pkg1 host1'.split())
        assert args.hosts == ['host1']
        assert args.install == "pkg1"

    def test_pkg_install_multiple_hosts(self):
        hostnames = ['host1', 'host2', 'host3']
        args = self.parser.parse_args('pkg --install pkg1'.split() + hostnames)
        assert args.hosts == hostnames
        assert args.install == "pkg1"

    def test_pkg_install_muliple_pkgs(self):
        args = self.parser.parse_args('pkg --install pkg1,pkg2 host1'.split())
        assert args.install == "pkg1,pkg2"

    def test_pkg_remove_host_required(self, capsys):
        with pytest.raises(SystemExit):
            self.parser.parse_args('pkg --remove pkg1'.split())
        out, err = capsys.readouterr()
        assert_too_few_arguments(err)

    def test_pkg_remove_one_host(self):
        args = self.parser.parse_args('pkg --remove pkg1 host1'.split())
        assert args.hosts == ['host1']
        assert args.remove == "pkg1"

    def test_pkg_remove_multiple_hosts(self):
        hostnames = ['host1', 'host2', 'host3']
        args = self.parser.parse_args('pkg --remove pkg1'.split() + hostnames)
        assert args.hosts == hostnames
        assert args.remove == "pkg1"

    def test_pkg_remove_muliple_pkgs(self):
        args = self.parser.parse_args('pkg --remove pkg1,pkg2 host1'.split())
        assert args.remove == "pkg1,pkg2"

    def test_pkg_install_remove_are_mutex(self, capsys):
        with pytest.raises(SystemExit):
            self.parser.parse_args('pkg --install pkg2 --remove pkg1 host1'.split())
        out, err = capsys.readouterr()
        assert "argument --remove: not allowed with argument --install" in err



"""
We deal (mostly) with remote hosts. To avoid special casing each different
commands (e.g. using `yum` as opposed to `apt`) we can make a one time call to
that remote host and set all the special cases for running commands depending
on the type of distribution/version we are dealing with.
"""
import logging
from ceph_deploy import exc
from ceph_deploy.util import versions
from ceph_deploy.hosts import debian, centos, fedora, suse, remotes, rhel
from ceph_deploy.connection import get_connection

logger = logging.getLogger()


def get(hostname,
        username=None,
        fallback=None,
        detect_sudo=True,
        use_rhceph=False,
        callbacks=None):
    """
    Retrieve the module that matches the distribution of a ``hostname``. This
    function will connect to that host and retrieve the distribution
    information, then return the appropriate module and slap a few attributes
    to that module defining the information it found from the hostname.

    For example, if host ``node1.example.com`` is an Ubuntu server, the
    ``debian`` module would be returned and the following would be set::

        module.name = 'ubuntu'
        module.release = '12.04'
        module.codename = 'precise'

    :param hostname: A hostname that is reachable/resolvable over the network
    :param fallback: Optional fallback to use if no supported distro is found
    :param use_rhceph: Whether or not to install RH Ceph on a RHEL machine or
                       the community distro.  Changes what host module is
                       returned for RHEL.
    :params callbacks: A list of callables that accept one argument (the actual
                       module that contains the connection) that will be
                       called, in order at the end of the instantiation of the
                       module.
    """
    conn = get_connection(
        hostname,
        username=username,
        logger=logging.getLogger(hostname),
        detect_sudo=detect_sudo
    )
    try:
        conn.import_module(remotes)
    except IOError as error:
        if 'already closed' in getattr(error, 'message', ''):
            raise RuntimeError('remote connection got closed, ensure ``requiretty`` is disabled for %s' % hostname)
    distro_name, release, codename = conn.remote_module.platform_information()
    if not codename or not _get_distro(distro_name):
        raise exc.UnsupportedPlatform(
            distro=distro_name,
            codename=codename,
            release=release)

    machine_type = conn.remote_module.machine_type()
    module = _get_distro(distro_name, use_rhceph=use_rhceph)
    module.name = distro_name
    module.normalized_name = _normalized_distro_name(distro_name)
    module.normalized_release = _normalized_release(release)
    module.distro = module.normalized_name
    module.is_el = module.normalized_name in ['redhat', 'centos', 'fedora', 'scientific', 'oracle', 'virtuozzo']
    module.is_rpm = module.normalized_name in ['redhat', 'centos',
                                               'fedora', 'scientific', 'suse', 'oracle', 'virtuozzo']
    module.is_deb = not module.is_rpm
    module.release = release
    module.codename = codename
    module.conn = conn
    module.machine_type = machine_type
    module.init = module.choose_init(module)
    module.packager = module.get_packager(module)
    # execute each callback if any
    if callbacks:
        for c in callbacks:
            c(module)
    return module


def _get_distro(distro, fallback=None, use_rhceph=False):
    if not distro:
        return

    distro = _normalized_distro_name(distro)
    distributions = {
        'debian': debian,
        'ubuntu': debian,
        'centos': centos,
        'scientific': centos,
        'oracle' : centos,
        'redhat': centos,
        'fedora': fedora,
        'suse': suse,
        'virtuozzo' : centos
        }

    if distro == 'redhat' and use_rhceph:
        return rhel
    else:
        return distributions.get(distro) or _get_distro(fallback)


def _normalized_distro_name(distro):
    distro = distro.lower()
    if distro.startswith(('redhat', 'red hat')):
        return 'redhat'
    elif distro.startswith(('scientific', 'scientific linux')):
        return 'scientific'
    elif distro.startswith('oracle'):
        return 'oracle'
    elif distro.startswith(('suse', 'opensuse')):
        return 'suse'
    elif distro.startswith('centos'):
        return 'centos'
    elif distro.startswith('linuxmint'):
        return 'ubuntu'
    elif distro.startswith('virtuozzo'):
        return 'virtuozzo'
    return distro


def _normalized_release(release):
    """
    A normalizer function to make sense of distro
    release versions.

    Returns an object with: major, minor, patch, and garbage

    These attributes can be accessed as ints with prefixed "int"
    attribute names, for example:

        normalized_version.int_major
    """
    # TODO: at some point deprecate this function so that we just
    # use this class directly (and update every test that calls it
    return versions.NormalizedVersion(release)



def uninstall(distro, purge=False):
    packages = [
        'ceph',
        'ceph-common',
        'ceph-mon',
        'ceph-osd',
        'ceph-radosgw'
        ]

    distro.packager.remove(packages)
    distro.packager.clean()



from ceph_deploy.util import paths
from ceph_deploy import conf
from ceph_deploy.lib import remoto
from ceph_deploy.util import constants
from ceph_deploy.util import system


def ceph_version(conn):
    """
    Log the remote ceph-version by calling `ceph --version`
    """
    return remoto.process.run(conn, ['ceph', '--version'])


def mon_create(distro, args, monitor_keyring):
    hostname = distro.conn.remote_module.shortname()
    logger = distro.conn.logger
    logger.debug('remote hostname: %s' % hostname)
    path = paths.mon.path(args.cluster, hostname)
    uid = distro.conn.remote_module.path_getuid(constants.base_path)
    gid = distro.conn.remote_module.path_getgid(constants.base_path)
    done_path = paths.mon.done(args.cluster, hostname)
    init_path = paths.mon.init(args.cluster, hostname, distro.init)

    conf_data = conf.ceph.load_raw(args)

    # write the configuration file
    distro.conn.remote_module.write_conf(
        args.cluster,
        conf_data,
        args.overwrite_conf,
    )

    # if the mon path does not exist, create it
    distro.conn.remote_module.create_mon_path(path, uid, gid)

    logger.debug('checking for done path: %s' % done_path)
    if not distro.conn.remote_module.path_exists(done_path):
        logger.debug('done path does not exist: %s' % done_path)
        if not distro.conn.remote_module.path_exists(paths.mon.constants.tmp_path):
            logger.info('creating tmp path: %s' % paths.mon.constants.tmp_path)
            distro.conn.remote_module.makedir(paths.mon.constants.tmp_path)
        keyring = paths.mon.keyring(args.cluster, hostname)

        logger.info('creating keyring file: %s' % keyring)
        distro.conn.remote_module.write_monitor_keyring(
            keyring,
            monitor_keyring,
            uid, gid,
        )

        user_args = []
        if uid != 0:
            user_args = user_args + [ '--setuser', str(uid) ]
        if gid != 0:
            user_args = user_args + [ '--setgroup', str(gid) ]

        remoto.process.run(
            distro.conn,
            [
                'ceph-mon',
                '--cluster', args.cluster,
                '--mkfs',
                '-i', hostname,
                '--keyring', keyring,
            ] + user_args
        )

        logger.info('unlinking keyring file %s' % keyring)
        distro.conn.remote_module.unlink(keyring)

    # create the done file
    distro.conn.remote_module.create_done_path(done_path, uid, gid)

    # create init path
    distro.conn.remote_module.create_init_path(init_path, uid, gid)

    # start mon service
    start_mon_service(distro, args.cluster, hostname)


def mon_add(distro, args, monitor_keyring):
    hostname = distro.conn.remote_module.shortname()
    logger = distro.conn.logger
    path = paths.mon.path(args.cluster, hostname)
    uid = distro.conn.remote_module.path_getuid(constants.base_path)
    gid = distro.conn.remote_module.path_getgid(constants.base_path)
    monmap_path = paths.mon.monmap(args.cluster, hostname)
    done_path = paths.mon.done(args.cluster, hostname)
    init_path = paths.mon.init(args.cluster, hostname, distro.init)

    conf_data = conf.ceph.load_raw(args)

    # write the configuration file
    distro.conn.remote_module.write_conf(
        args.cluster,
        conf_data,
        args.overwrite_conf,
    )

    # if the mon path does not exist, create it
    distro.conn.remote_module.create_mon_path(path, uid, gid)

    logger.debug('checking for done path: %s' % done_path)
    if not distro.conn.remote_module.path_exists(done_path):
        logger.debug('done path does not exist: %s' % done_path)
        if not distro.conn.remote_module.path_exists(paths.mon.constants.tmp_path):
            logger.info('creating tmp path: %s' % paths.mon.constants.tmp_path)
            distro.conn.remote_module.makedir(paths.mon.constants.tmp_path)
        keyring = paths.mon.keyring(args.cluster, hostname)

        logger.info('creating keyring file: %s' % keyring)
        distro.conn.remote_module.write_monitor_keyring(
            keyring,
            monitor_keyring,
            uid, gid,
        )

        # get the monmap
        remoto.process.run(
            distro.conn,
            [
                'ceph',
                '--cluster', args.cluster,
                'mon',
                'getmap',
                '-o',
                monmap_path,
            ],
        )

        # now use it to prepare the monitor's data dir
        user_args = []
        if uid != 0:
            user_args = user_args + [ '--setuser', str(uid) ]
        if gid != 0:
            user_args = user_args + [ '--setgroup', str(gid) ]

        remoto.process.run(
            distro.conn,
            [
                'ceph-mon',
                '--cluster', args.cluster,
                '--mkfs',
                '-i', hostname,
                '--monmap',
                monmap_path,
                '--keyring', keyring,
            ] + user_args
        )

        logger.info('unlinking keyring file %s' % keyring)
        distro.conn.remote_module.unlink(keyring)

    # create the done file
    distro.conn.remote_module.create_done_path(done_path, uid, gid)

    # create init path
    distro.conn.remote_module.create_init_path(init_path, uid, gid)

    # start mon service
    start_mon_service(distro, args.cluster, hostname)


def map_components(notsplit_packages, components):
    """
    Returns a list of packages to install based on component names

    This is done by checking if a component is in notsplit_packages,
    if it is, we know we need to install 'ceph' instead of the
    raw component name.  Essentially, this component hasn't been
    'split' from the master 'ceph' package yet.
    """
    packages = set()

    for c in components:
        if c in notsplit_packages:
            packages.add('ceph')
        else:
            packages.add(c)

    return list(packages)


def start_mon_service(distro, cluster, hostname):
    """
    start mon service depending on distro init
    """
    if distro.init == 'sysvinit':
        service = distro.conn.remote_module.which_service()
        remoto.process.run(
            distro.conn,
            [
                service,
                'ceph',
                '-c',
                '/etc/ceph/{cluster}.conf'.format(cluster=cluster),
                'start',
                'mon.{hostname}'.format(hostname=hostname)
            ],
            timeout=7,
        )
        system.enable_service(distro.conn)

    elif distro.init == 'upstart':
        remoto.process.run(
             distro.conn,
             [
                 'initctl',
                 'emit',
                 'ceph-mon',
                 'cluster={cluster}'.format(cluster=cluster),
                 'id={hostname}'.format(hostname=hostname),
             ],
             timeout=7,
         )

    elif distro.init == 'systemd':
       # enable ceph target for this host (in case it isn't already enabled)
        remoto.process.run(
            distro.conn,
            [
                'systemctl',
                'enable',
                'ceph.target'
            ],
            timeout=7,
        )

        # enable and start this mon instance
        remoto.process.run(
            distro.conn,
            [
                'systemctl',
                'enable',
                'ceph-mon@{hostname}'.format(hostname=hostname),
            ],
            timeout=7,
        )
        remoto.process.run(
            distro.conn,
            [
                'systemctl',
                'start',
                'ceph-mon@{hostname}'.format(hostname=hostname),
            ],
            timeout=7,
        )



from ceph_deploy.exc import ExecutableNotFound
from ceph_deploy.lib import remoto


def executable_path(conn, executable):
    """
    Remote validator that accepts a connection object to ensure that a certain
    executable is available returning its full path if so.

    Otherwise an exception with thorough details will be raised, informing the
    user that the executable was not found.
    """
    executable_path = conn.remote_module.which(executable)
    if not executable_path:
        raise ExecutableNotFound(executable, conn.hostname)
    return executable_path


def is_systemd(conn):
    """
    Attempt to detect if a remote system is a systemd one or not
    by looking into ``/proc`` just like the ceph init script does::

        # detect systemd
        # SYSTEMD=0
        grep -qs systemd /proc/1/comm && SYSTEMD=1
    """
    return conn.remote_module.grep(
        'systemd',
        '/proc/1/comm'
    )


def is_upstart(conn):
    """
    This helper should only used as a fallback (last resort) as it is not
    guaranteed that it will be absolutely correct.
    """
    # it may be possible that we may be systemd and the caller never checked
    # before so lets do that
    if is_systemd(conn):
        return False

    # get the initctl executable, if it doesn't exist we can't proceed so we
    # are probably not upstart
    initctl = conn.remote_module.which('initctl')
    if not initctl:
        return False

    # finally, try and get output from initctl that might hint this is an upstart
    # system. On a Ubuntu 14.04.2 system this would look like:
    # $ initctl version
    # init (upstart 1.12.1)
    stdout, stderr, _ = remoto.process.check(
        conn,
        [initctl, 'version'],
    )
    result_string = b' '.join(stdout)
    if b'upstart' in result_string:
        return True
    return False


def enable_service(conn, service='ceph'):
    """
    Enable a service on a remote host depending on the type of init system.
    Obviously, this should be done for RHEL/Fedora/CentOS systems.

    This function does not do any kind of detection.
    """
    if is_systemd(conn):
        remoto.process.run(
            conn,
            [
                'systemctl',
                'enable',
                '{service}'.format(service=service),
            ]
        )
    else:
        remoto.process.run(
            conn,
            [
                'chkconfig',
                '{service}'.format(service=service),
                'on',
            ]
        )



"""
This module is meant for vendorizing Python libraries. Most libraries will need
to have some ``sys.path`` alterations done unless they are doing relative
imports.

Do **not** add anything to this module that does not represent a vendorized
library.

Vendored libraries should go into the ``vendor`` directory and imported from
there. This is so we allow libraries that are installed normally to be imported
if the vendored module is not available.

The import dance here is done so that all other imports throught ceph-deploy
are kept the same regardless of where the module comes from.

The expected way to import remoto would look like this::

    from ceph_deploy.lib import remoto

"""

try:
    # vendored
    from .vendor import remoto
except ImportError:
    # normally installed
    import remoto  # noqa



from . import mon  # noqa
from .install import install, mirror_install, repo_install  # noqa
from .uninstall import uninstall  # noqa
import logging

from ceph_deploy.util import pkg_managers

# Allow to set some information about this distro
#

log = logging.getLogger(__name__)

distro = None
release = None
codename = None

def choose_init(module):
    """
    Select a init system

    Returns the name of a init system (upstart, sysvinit ...).
    """
    init_mapping = { '11' : 'sysvinit', # SLE_11
        '12' : 'systemd',               # SLE_12
        '13.1' : 'systemd',             # openSUSE_13.1
        }
    return init_mapping.get(release, 'systemd')


def get_packager(module):
    return pkg_managers.Zypper(module)




def mon_hosts(mons):
    """
    Iterate through list of MON hosts, return tuples of (name, host).
    """
    for m in mons:
        if m.count(':'):
            (name, host) = m.split(':')
        else:
            name = m
            host = m
            if name.count('.') > 0:
                name = name.split('.')[0]
        yield (name, host)

def remote_shortname(socket):
    """
    Obtains remote hostname of the socket and cuts off the domain part
    of its FQDN.
    """
    return socket.gethostname().split('.', 1)[0]




import pytest

from ceph_deploy.cli import get_parser
from ceph_deploy.tests.util import assert_too_few_arguments

SUBCMDS_WITH_ARGS = ['push', 'pull']


class TestParserConfig(object):

    def setup(self):
        self.parser = get_parser()

    def test_config_help(self, capsys):
        with pytest.raises(SystemExit):
            self.parser.parse_args('config --help'.split())
        out, err = capsys.readouterr()
        assert 'usage: ceph-deploy config' in out
        assert 'positional arguments:' in out
        assert 'optional arguments:' in out

    @pytest.mark.parametrize('cmd', SUBCMDS_WITH_ARGS)
    def test_config_subcommands_with_args(self, cmd):
        self.parser.parse_args(['config'] + ['%s' % cmd] + ['host1'])

    def test_config_invalid_subcommand(self, capsys):
        with pytest.raises(SystemExit):
            self.parser.parse_args('config bork'.split())
        out, err = capsys.readouterr()
        assert 'invalid choice' in err

    def test_config_push_host_required(self, capsys):
        with pytest.raises(SystemExit):
            self.parser.parse_args('config push'.split())
        out, err = capsys.readouterr()
        assert_too_few_arguments(err)

    def test_config_push_one_host(self):
        args = self.parser.parse_args('config push host1'.split())
        assert args.client == ['host1']

    def test_config_push_multiple_hosts(self):
        hostnames = ['host1', 'host2', 'host3']
        args = self.parser.parse_args('config push'.split() + hostnames)
        assert args.client == hostnames

    def test_config_pull_host_required(self, capsys):
        with pytest.raises(SystemExit):
            self.parser.parse_args('config pull'.split())
        out, err = capsys.readouterr()
        assert_too_few_arguments(err)

    def test_config_pull_one_host(self):
        args = self.parser.parse_args('config pull host1'.split())
        assert args.client == ['host1']

    def test_config_pull_multiple_hosts(self):
        hostnames = ['host1', 'host2', 'host3']
        args = self.parser.parse_args('config pull'.split() + hostnames)
        assert args.client == hostnames



import socket
from contextlib import closing

#sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
#sock.settimeout(2)  # 2 Second Timeout


def generate_ips(start_ip, end_ip):
    start = list(map(int, start_ip.split(".")))
    end = list(map(int, end_ip.split(".")))
    temp = start
    ip_range = []

    ip_range.append(start_ip)
    while temp != end:
        start[3] += 1
        for i in (3, 2, 1):
            if temp[i] == 256:
                temp[i] = 0
                temp[i-1] += 1
        ip_range.append(".".join(map(str, temp)))

    return ip_range

count = 0
for address in generate_ips('192.168.1.1', '192.168.1.255'):
    count += 1
    for port in [80, 7480]:
        print address, port
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.settimeout(2)  # 2 Second Timeout

        result = sock.connect_ex((address, port))
        if result == 0:
            print 'port %s OPEN on %s' % (port, address)
        sock.close()

print 'Total scanned addresses: %s' % count



"""
Common paths for mon, based on the constant file paths defined in
``ceph_deploy.util.constants``.
All functions return a string representation of the absolute path
construction.
"""
from os.path import join

from ceph_deploy.util import constants


def base(cluster):
    cluster = "%s-" % cluster
    return join(constants.mon_path, cluster)


def path(cluster, hostname):
    """
    Example usage::

        >>> from ceph_deploy.util.paths import mon
        >>> mon.path('mycluster', 'hostname')
        /var/lib/ceph/mon/mycluster-myhostname
    """
    return "%s%s" % (base(cluster), hostname)


def done(cluster, hostname):
    """
    Example usage::

        >>> from ceph_deploy.util.paths import mon
        >>> mon.done('mycluster', 'hostname')
        /var/lib/ceph/mon/mycluster-myhostname/done
    """
    return join(path(cluster, hostname), 'done')


def init(cluster, hostname, init):
    """
    Example usage::

        >>> from ceph_deploy.util.paths import mon
        >>> mon.init('mycluster', 'hostname', 'init')
        /var/lib/ceph/mon/mycluster-myhostname/init
    """
    return join(path(cluster, hostname), init)


def keyring(cluster, hostname):
    """
    Example usage::

        >>> from ceph_deploy.util.paths import mon
        >>> mon.keyring('mycluster', 'myhostname')
        /var/lib/ceph/tmp/mycluster-myhostname.mon.keyring
    """
    keyring_file = '%s-%s.mon.keyring' % (cluster, hostname)
    return join(constants.tmp_path, keyring_file)


def asok(cluster, hostname):
    """
    Example usage::

        >>> from ceph_deploy.util.paths import mon
        >>> mon.asok('mycluster', 'myhostname')
        /var/run/ceph/mycluster-mon.myhostname.asok
    """
    asok_file = '%s-mon.%s.asok' % (cluster, hostname)
    return join(constants.base_run_path, asok_file)


def monmap(cluster, hostname):
    """
    Example usage::

        >>> from ceph_deploy.util.paths import mon
        >>> mon.monmap('mycluster', 'myhostname')
        /var/lib/ceph/tmp/mycluster.myhostname.monmap
    """
    monmap
    mon_map_file = '%s.%s.monmap' % (cluster, hostname)
    return join(constants.tmp_path, mon_map_file)



"""
gateway code for initiating popen, socket and ssh connections.
(c) 2004-2013, Holger Krekel and others
"""

import sys, os, inspect, types, linecache
import textwrap
import execnet
from execnet.gateway_base import Message
from execnet import gateway_base
importdir = os.path.dirname(os.path.dirname(execnet.__file__))

class Gateway(gateway_base.BaseGateway):
    """ Gateway to a local or remote Python Intepreter. """

    def __init__(self, io, spec):
        super(Gateway, self).__init__(io=io, id=spec.id, _startcount=1)
        self.spec = spec
        self._initreceive()

    @property
    def remoteaddress(self):
        return self._io.remoteaddress

    def __repr__(self):
        """ return string representing gateway type and status. """
        try:
            r = (self.hasreceiver() and 'receive-live' or 'not-receiving')
            i = len(self._channelfactory.channels())
        except AttributeError:
            r = "uninitialized"
            i = "no"
        return "<%s id=%r %s, %s model, %s active channels>" %(
                self.__class__.__name__, self.id, r, self.execmodel.backend, i)

    def exit(self):
        """ trigger gateway exit.  Defer waiting for finishing
        of receiver-thread and subprocess activity to when
        group.terminate() is called.
        """
        self._trace("gateway.exit() called")
        if self not in self._group:
            self._trace("gateway already unregistered with group")
            return
        self._group._unregister(self)
        try:
            self._trace("--> sending GATEWAY_TERMINATE")
            self._send(Message.GATEWAY_TERMINATE)
            self._trace("--> io.close_write")
            self._io.close_write()
        except (ValueError, EOFError, IOError):
            v = sys.exc_info()[1]
            self._trace("io-error: could not send termination sequence")
            self._trace(" exception: %r" % v)

    def reconfigure(self, py2str_as_py3str=True, py3str_as_py2str=False):
        """
        set the string coercion for this gateway
        the default is to try to convert py2 str as py3 str,
        but not to try and convert py3 str to py2 str
        """
        self._strconfig = (py2str_as_py3str, py3str_as_py2str)
        data = gateway_base.dumps_internal(self._strconfig)
        self._send(Message.RECONFIGURE, data=data)


    def _rinfo(self, update=False):
        """ return some sys/env information from remote. """
        if update or not hasattr(self, '_cache_rinfo'):
            ch = self.remote_exec(rinfo_source)
            self._cache_rinfo = RInfo(ch.receive())
        return self._cache_rinfo

    def hasreceiver(self):
        """ return True if gateway is able to receive data. """
        return self._receivepool.active_count() > 0

    def remote_status(self):
        """ return information object about remote execution status. """
        channel = self.newchannel()
        self._send(Message.STATUS, channel.id)
        statusdict = channel.receive()
        # the other side didn't actually instantiate a channel
        # so we just delete the internal id/channel mapping
        self._channelfactory._local_close(channel.id)
        return RemoteStatus(statusdict)

    def remote_exec(self, source, **kwargs):
        """ return channel object and connect it to a remote
            execution thread where the given ``source`` executes.

            * ``source`` is a string: execute source string remotely
              with a ``channel`` put into the global namespace.
            * ``source`` is a pure function: serialize source and
              call function with ``**kwargs``, adding a
              ``channel`` object to the keyword arguments.
            * ``source`` is a pure module: execute source of module
              with a ``channel`` in its global namespace

            In all cases the binding ``__name__='__channelexec__'``
            will be available in the global namespace of the remotely
            executing code.
        """
        call_name = None
        if isinstance(source, types.ModuleType):
            linecache.updatecache(inspect.getsourcefile(source))
            source = inspect.getsource(source)
        elif isinstance(source, types.FunctionType):
            call_name = source.__name__
            source = _source_of_function(source)
        else:
            source = textwrap.dedent(str(source))

        if call_name is None and kwargs:
            raise TypeError("can't pass kwargs to non-function remote_exec")

        channel = self.newchannel()
        self._send(Message.CHANNEL_EXEC,
                   channel.id,
                   gateway_base.dumps_internal((source, call_name, kwargs)))
        return channel

    def remote_init_threads(self, num=None):
        """ DEPRECATED.  Is currently a NO-OPERATION already."""
        print ("WARNING: remote_init_threads() is a no-operation in execnet-1.2")

class RInfo:
    def __init__(self, kwargs):
        self.__dict__.update(kwargs)

    def __repr__(self):
        info = ", ".join(["%s=%s" % item
                for item in self.__dict__.items()])
        return "<RInfo %r>" % info

RemoteStatus = RInfo

def rinfo_source(channel):
    import sys, os
    channel.send(dict(
        executable = sys.executable,
        version_info = sys.version_info[:5],
        platform = sys.platform,
        cwd = os.getcwd(),
        pid = os.getpid(),
    ))


def _find_non_builtin_globals(source, codeobj):
    try:
        import ast
    except ImportError:
        return None
    try:
        import __builtin__
    except ImportError:
        import builtins as __builtin__

    vars = dict.fromkeys(codeobj.co_varnames)
    all = []
    for node in ast.walk(ast.parse(source)):
        if (isinstance(node, ast.Name) and node.id not in vars and
            node.id not in __builtin__.__dict__):
            all.append(node.id)
    return all


def _source_of_function(function):
    if function.__name__ == '<lambda>':
        raise ValueError("can't evaluate lambda functions'")
    #XXX: we dont check before remote instanciation
    #     if arguments are used propperly
    args, varargs, keywords, defaults = inspect.getargspec(function)
    if args[0] != 'channel':
        raise ValueError('expected first function argument to be `channel`')

    if sys.version_info < (3,0):
        closure = function.func_closure
        codeobj = function.func_code
    else:
        closure = function.__closure__
        codeobj = function.__code__

    if closure is not None:
        raise ValueError("functions with closures can't be passed")

    try:
        source = inspect.getsource(function)
    except IOError:
        raise ValueError("can't find source file for %s" % function)

    source = textwrap.dedent(source) # just for inner functions

    used_globals = _find_non_builtin_globals(source, codeobj)
    if used_globals:
        raise ValueError(
            "the use of non-builtin globals isn't supported",
            used_globals,
        )

    return source




import socket
from ceph_deploy.lib import remoto


def get_connection(hostname, username, logger, threads=5, use_sudo=None, detect_sudo=True):
    """
    A very simple helper, meant to return a connection
    that will know about the need to use sudo.
    """
    if username:
        hostname = "%s@%s" % (username, hostname)
    try:
        conn = remoto.Connection(
            hostname,
            logger=logger,
            threads=threads,
            detect_sudo=detect_sudo,
        )

        # Set a timeout value in seconds to disconnect and move on
        # if no data is sent back.
        conn.global_timeout = 300
        logger.debug("connected to host: %s " % hostname)
        return conn

    except Exception as error:
        msg = "connecting to host: %s " % hostname
        errors = "resulted in errors: %s %s" % (error.__class__.__name__, error)
        raise RuntimeError(msg + errors)


def get_local_connection(logger, use_sudo=False):
    """
    Helper for local connections that are sometimes needed to operate
    on local hosts
    """
    return get_connection(
        socket.gethostname(),  # cannot rely on 'localhost' here
        None,
        logger=logger,
        threads=1,
        use_sudo=use_sudo,
        detect_sudo=False
    )





def admin_command(sudo, command):
    """
    If sudo is needed, make sure the command is prepended
    correctly, otherwise return the command as it came.

    :param sudo: A boolean representing the intention of having a sudo command
                (or not)
    :param command: A list of the actual command to execute with Popen.
    """
    if sudo:
        if not isinstance(command, list):
            command = [command]
        return ['sudo'] + [cmd for cmd in command]
    return command


class RemoteError(object):

    def __init__(self, traceback):
        self.orig_traceback = traceback
        self.exception_line = ''
        self.exception_name = self.get_exception_name()

    def get_exception_name(self):
        for tb_line in reversed(self.orig_traceback.split('\n')):
            if tb_line:
                for word in tb_line.split():
                    if word.endswith(':'):  # exception!
                        self.exception_line = tb_line
                        return word.strip().strip(':')



import pytest

from ceph_deploy.cli import get_parser
from ceph_deploy.tests.util import assert_too_few_arguments

SUBCMDS_WITH_ARGS = ['list', 'prepare', 'activate', 'zap']


class TestParserDisk(object):

    def setup(self):
        self.parser = get_parser()

    def test_disk_help(self, capsys):
        with pytest.raises(SystemExit):
            self.parser.parse_args('disk --help'.split())
        out, err = capsys.readouterr()
        assert 'usage: ceph-deploy disk' in out
        assert 'positional arguments:' in out
        assert 'optional arguments:' in out

    @pytest.mark.parametrize('cmd', SUBCMDS_WITH_ARGS)
    def test_disk_valid_subcommands_with_args(self, cmd):
        self.parser.parse_args(['disk'] + ['%s' % cmd] + ['host1'])

    def test_disk_invalid_subcommand(self, capsys):
        with pytest.raises(SystemExit):
            self.parser.parse_args('disk bork'.split())
        out, err = capsys.readouterr()
        assert 'invalid choice' in err

    def test_disk_list_help(self, capsys):
        with pytest.raises(SystemExit):
            self.parser.parse_args('disk list --help'.split())
        out, err = capsys.readouterr()
        assert 'usage: ceph-deploy disk list' in out

    def test_disk_list_host_required(self, capsys):
        with pytest.raises(SystemExit):
            self.parser.parse_args('disk list'.split())
        out, err = capsys.readouterr()
        assert_too_few_arguments(err)

    def test_disk_list_single_host(self):
        args = self.parser.parse_args('disk list host1'.split())
        assert args.disk[0][0] == 'host1'

    def test_disk_list_multi_host(self):
        hostnames = ['host1', 'host2', 'host3']
        args = self.parser.parse_args('disk list'.split() + hostnames)
        # args.disk is a list of tuples, and tuple[0] is the hostname
        hosts = [x[0] for x in args.disk]
        assert hosts == hostnames

    def test_disk_prepare_help(self, capsys):
        with pytest.raises(SystemExit):
            self.parser.parse_args('disk prepare --help'.split())
        out, err = capsys.readouterr()
        assert 'usage: ceph-deploy disk prepare' in out

    def test_disk_prepare_zap_default_false(self):
        args = self.parser.parse_args('disk prepare host1:sdb'.split())
        assert args.zap_disk is False

    def test_disk_prepare_zap_true(self):
        args = self.parser.parse_args('disk prepare --zap-disk host1:sdb'.split())
        assert args.zap_disk is True

    def test_disk_prepare_fstype_default_xfs(self):
        args = self.parser.parse_args('disk prepare host1:sdb'.split())
        assert args.fs_type == "xfs"

    def test_disk_prepare_fstype_btrfs(self):
        args = self.parser.parse_args('disk prepare --fs-type btrfs host1:sdb'.split())
        assert args.fs_type == "btrfs"

    def test_disk_prepare_fstype_invalid(self, capsys):
        with pytest.raises(SystemExit):
            self.parser.parse_args('disk prepare --fs-type bork host1:sdb'.split())
        out, err = capsys.readouterr()
        assert 'invalid choice' in err

    def test_disk_prepare_dmcrypt_default_false(self):
        args = self.parser.parse_args('disk prepare host1:sdb'.split())
        assert args.dmcrypt is False

    def test_disk_prepare_dmcrypt_true(self):
        args = self.parser.parse_args('disk prepare --dmcrypt host1:sdb'.split())
        assert args.dmcrypt is True

    def test_disk_prepare_dmcrypt_key_dir_default(self):
        args = self.parser.parse_args('disk prepare host1:sdb'.split())
        assert args.dmcrypt_key_dir == "/etc/ceph/dmcrypt-keys"

    def test_disk_prepare_dmcrypt_key_dir_custom(self):
        args = self.parser.parse_args('disk prepare --dmcrypt --dmcrypt-key-dir /tmp/keys host1:sdb'.split())
        assert args.dmcrypt_key_dir == "/tmp/keys"

    def test_disk_prepare_host_required(self, capsys):
        with pytest.raises(SystemExit):
            self.parser.parse_args('disk prepare'.split())
        out, err = capsys.readouterr()
        assert_too_few_arguments(err)

    def test_disk_prepare_single_host(self):
        args = self.parser.parse_args('disk prepare host1:sdb'.split())
        assert args.disk[0][0] == 'host1'

    def test_disk_prepare_multi_host(self):
        hostnames = ['host1', 'host2', 'host3']
        args = self.parser.parse_args('disk prepare'.split() + [x + ":sdb" for x in hostnames])
        # args.disk is a list of tuples, and tuple[0] is the hostname
        hosts = [x[0] for x in args.disk]
        assert hosts == hostnames

    def test_disk_activate_help(self, capsys):
        with pytest.raises(SystemExit):
            self.parser.parse_args('disk activate --help'.split())
        out, err = capsys.readouterr()
        assert 'usage: ceph-deploy disk activate' in out

    def test_disk_activate_host_required(self, capsys):
        with pytest.raises(SystemExit):
            self.parser.parse_args('disk activate'.split())
        out, err = capsys.readouterr()
        assert_too_few_arguments(err)

    def test_disk_activate_single_host(self):
        args = self.parser.parse_args('disk activate host1:sdb1'.split())
        assert args.disk[0][0] == 'host1'

    def test_disk_activate_multi_host(self):
        hostnames = ['host1', 'host2', 'host3']
        args = self.parser.parse_args('disk activate'.split() + [x + ":sdb1" for x in hostnames])
        # args.disk is a list of tuples, and tuple[0] is the hostname
        hosts = [x[0] for x in args.disk]
        assert hosts == hostnames

    def test_disk_zap_help(self, capsys):
        with pytest.raises(SystemExit):
            self.parser.parse_args('disk zap --help'.split())
        out, err = capsys.readouterr()
        assert 'usage: ceph-deploy disk zap' in out

    def test_disk_zap_host_required(self, capsys):
        with pytest.raises(SystemExit):
            self.parser.parse_args('disk zap'.split())
        out, err = capsys.readouterr()
        assert_too_few_arguments(err)

    def test_disk_zap_single_host(self):
        args = self.parser.parse_args('disk zap host1:sdb'.split())
        assert args.disk[0][0] == 'host1'

    def test_disk_zap_multi_host(self):
        hostnames = ['host1', 'host2', 'host3']
        args = self.parser.parse_args('disk zap'.split() + [x + ":sdb" for x in hostnames])
        # args.disk is a list of tuples, and tuple[0] is the hostname
        hosts = [x[0] for x in args.disk]
        assert hosts == hostnames



import argparse
import re


ALPHANUMERIC_RE = re.compile(r'^[a-zA-Z][a-zA-Z0-9]*$')


def alphanumeric(s):
    """
    Enforces string to be alphanumeric with leading alpha.
    """
    if not ALPHANUMERIC_RE.match(s):
        raise argparse.ArgumentTypeError(
            'argument must start with a letter and contain only letters and numbers',
            )
    return s



from setuptools import setup, find_packages
import os
import sys
import ceph_deploy
from vendor import vendorize, clean_vendor


def read(fname):
    path = os.path.join(os.path.dirname(__file__), fname)
    f = open(path)
    return f.read()

install_requires = []
pyversion = sys.version_info[:2]
if pyversion < (2, 7) or (3, 0) <= pyversion <= (3, 1):
    install_requires.append('argparse')

#
# Add libraries that are not part of install_requires but only if we really
# want to, specified by the environment flag
#

if os.environ.get('CEPH_DEPLOY_NO_VENDOR'):
    clean_vendor('remoto')
else:
    vendorize([
        ('remoto', '0.0.29', ['python', 'vendor.py']),
    ])


setup(
    name='ceph-deploy',
    version=ceph_deploy.__version__,
    packages=find_packages(),

    author='Inktank',
    author_email='ceph-devel@vger.kernel.org',
    description='Deploy Ceph with minimal infrastructure',
    long_description=read('README.rst'),
    license='MIT',
    keywords='ceph deploy',
    url="https://github.com/ceph/ceph-deploy",

    install_requires=[
        'setuptools',
        ] + install_requires,

    tests_require=[
        'pytest >=2.1.3',
        'mock >=1.0b1',
        ],

    entry_points={

        'console_scripts': [
            'ceph-deploy = ceph_deploy.cli:main',
            ],

        'ceph_deploy.cli': [
            'new = ceph_deploy.new:make',
            'install = ceph_deploy.install:make',
            'uninstall = ceph_deploy.install:make_uninstall',
            'purge = ceph_deploy.install:make_purge',
            'purgedata = ceph_deploy.install:make_purge_data',
            'mon = ceph_deploy.mon:make',
            'gatherkeys = ceph_deploy.gatherkeys:make',
            'osd = ceph_deploy.osd:make',
            'disk = ceph_deploy.osd:make_disk',
            'mds = ceph_deploy.mds:make',
            'mgr = ceph_deploy.mgr:make',
            'forgetkeys = ceph_deploy.forgetkeys:make',
            'config = ceph_deploy.config:make',
            'admin = ceph_deploy.admin:make',
            'pkg = ceph_deploy.pkg:make',
            'calamari = ceph_deploy.calamari:make',
            'rgw = ceph_deploy.rgw:make',
            'repo = ceph_deploy.repo:make',
            ],

        },
    )



import logging
from . import hosts


LOG = logging.getLogger(__name__)


def install(args):
    packages = args.install.split(',')
    for hostname in args.hosts:
        distro = hosts.get(hostname, username=args.username)
        LOG.info(
            'Distro info: %s %s %s',
            distro.name,
            distro.release,
            distro.codename
        )
        rlogger = logging.getLogger(hostname)
        rlogger.info('installing packages on %s' % hostname)
        # Do not timeout on package install. If you we this command to install
        # e.g. ceph-selinux or some other package with long post script we can
        # easily timeout in the 5 minutes that we use as a default timeout,
        # turning off the timeout completely for the time we run the command
        # should make this much more safe.
        distro.conn.global_timeout = None
        distro.packager.install(packages)
        distro.conn.exit()


def remove(args):
    packages = args.remove.split(',')
    for hostname in args.hosts:
        distro = hosts.get(hostname, username=args.username)
        LOG.info(
            'Distro info: %s %s %s',
            distro.name,
            distro.release,
            distro.codename
        )

        rlogger = logging.getLogger(hostname)
        rlogger.info('removing packages from %s' % hostname)
        # Do not timeout on package removal. If we use this command to remove
        # e.g. ceph-selinux or some other package with long post script we can
        # easily timeout in the 5 minutes that we use as a default timeout,
        # turning off the timeout completely for the time we run the command
        # should make this much more safe.
        distro.conn.global_timeout = None
        distro.packager.remove(packages)
        distro.conn.exit()


def pkg(args):
    if args.install:
        install(args)
    elif args.remove:
        remove(args)


def make(parser):
    """
    Manage packages on remote hosts.
    """

    action = parser.add_mutually_exclusive_group()

    action.add_argument(
        '--install',
        metavar='PKG(s)',
        help='Comma-separated package(s) to install',
    )

    action.add_argument(
        '--remove',
        metavar='PKG(s)',
        help='Comma-separated package(s) to remove',
    )

    parser.add_argument(
        'hosts',
        nargs='+',
    )

    parser.set_defaults(
        func=pkg,
    )



import argparse


class ToggleRawTextHelpFormatter(argparse.HelpFormatter):
    """ArgParse help formatter that allows raw text in individual help strings

        Inspired by the SmartFormatter at
           https://bitbucket.org/ruamel/std.argparse

       Normally to include newlines in the help output of argparse, you have
       use argparse.RawDescriptionHelpFormatter. But this means raw text is enabled
       everywhere, and not just for specific help entries where you might need it.

       This help formatter allows for you to optional enable/toggle raw text on
       individual menu items by prefixing the help string with 'R|'.

       Example:

       parser.formatter_class = ToggleRawTextHelpFormatter
       parser.add_argument('--verbose', action=store_true,
                           help='Enable verbose mode')
       #Above help is formatted just as default argparse.HelpFormatter

       parser.add_argument('--complex-arg', action=store_true,
                           help=('R|This help description use '
                                 'newlines and tabs and they will be preserved in'
                                 'the help output.\n\n'
                                 '\tHow cool is that?'))
    """
    def _split_lines(self, text, width):
        if text.startswith('R|'):
            return text[2:].splitlines()  
        return argparse.HelpFormatter._split_lines(self, text, width)



from execnet.gateway_bootstrap import HostNotFound
import sys

try: bytes
except NameError: bytes = str

class SocketIO:
    def __init__(self, sock, execmodel):
        self.sock = sock
        self.execmodel = execmodel
        socket = execmodel.socket
        try:
            sock.setsockopt(socket.SOL_IP, socket.IP_TOS, 0x10)# IPTOS_LOWDELAY
            sock.setsockopt(socket.SOL_TCP, socket.TCP_NODELAY, 1)
        except (AttributeError, socket.error):
            sys.stderr.write("WARNING: cannot set socketoption")

    def read(self, numbytes):
        "Read exactly 'bytes' bytes from the socket."
        buf = bytes()
        while len(buf) < numbytes:
            t = self.sock.recv(numbytes - len(buf))
            if not t:
                raise EOFError
            buf += t
        return buf

    def write(self, data):
        self.sock.sendall(data)

    def close_read(self):
        try:
            self.sock.shutdown(0)
        except self.execmodel.socket.error:
            pass
    def close_write(self):
        try:
            self.sock.shutdown(1)
        except self.execmodel.socket.error:
            pass

    def wait(self):
        pass

    def kill(self):
        pass


def start_via(gateway, hostport=None):
    """ return a host, port tuple,
        after instanciating a socketserver on the given gateway
    """
    if hostport is None:
        host, port = ('localhost', 0)
    else:
        host, port = hostport

    from execnet.script import socketserver

    # execute the above socketserverbootstrap on the other side
    channel = gateway.remote_exec(socketserver)
    channel.send((host, port))
    (realhost, realport) = channel.receive()
    #self._trace("new_remote received"
    #               "port=%r, hostname = %r" %(realport, hostname))
    if not realhost or realhost=="0.0.0.0":
        realhost = "localhost"
    return realhost, realport


def create_io(spec, group, execmodel):
    assert not spec.python, (
        "socket: specifying python executables not yet supported")
    gateway_id = spec.installvia
    if gateway_id:
        host, port = start_via(group[gateway_id])
    else:
        host, port = spec.socket.split(":")
        port = int(port)

    socket = execmodel.socket
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    io = SocketIO(sock, execmodel)
    io.remoteaddress = '%s:%d' % (host, port)
    try:
        sock.connect((host, port))
    except execmodel.socket.gaierror:
        raise HostNotFound(str(sys.exc_info()[1]))
    return io
